from collections.abc import Callable
from functools import singledispatchmethod
from typing import Any

from lionabc import BaseManager
from lionfuncs import to_dict, to_list

from lion_core.action.function_calling import FunctionCalling
from lion_core.action.tool import Tool, func_to_tool
from lion_core.communication.action_request import ActionRequest

REGISTERABLE_TOOL = Tool | Callable[..., Any]
FINDABLE_TOOL = REGISTERABLE_TOOL | str
INPUTTABLE_TOOL = dict[str, Any] | bool | FINDABLE_TOOL
TOOL_TYPE = FINDABLE_TOOL | list[FINDABLE_TOOL] | INPUTTABLE_TOOL


class ToolManager(BaseManager):
    """
    Manages tools in the system.

    Responsible for managing a registry of tools, which can be functions or
    callable objects. Provides methods to register tools, match function
    calls to tools, and invoke tools asynchronously.

    Attributes:
        registry: Mapping of tool names to Tool objects.
    """

    def __init__(self, registry: dict[str, Tool] | None = None) -> None:
        """Initialize the ToolManager instance.

        Args:
            registry: Optional dictionary of pre-registered tools.
                      Keys are tool names, values are Tool objects.
        """
        self.registry: dict[str, Tool] = registry or {}

    def __contains__(self, tool: FINDABLE_TOOL) -> bool:
        """Check if a tool is registered in the registry.

        Args:
            tool: The tool to check. Can be a Tool object, a string
                  (tool name), or a callable.

        Returns:
            True if the tool is registered, False otherwise.
        """
        if isinstance(tool, Tool):
            return tool.function_name in self.registry
        elif isinstance(tool, str):
            return tool in self.registry
        elif callable(tool):
            return tool.__name__ in self.registry
        return False

    def register_tool(
        self,
        tool: REGISTERABLE_TOOL,
        update: bool = False,
    ) -> None:
        """Register a single tool in the registry.

        Args:
            tool: The tool to register. Can be a Tool object or a callable.
            update: If True, update the tool if it's already registered.
                    If False, raise an error if already registered.

        Raises:
            ValueError: If tool is already registered and update is False.
            TypeError: If the provided tool is not a Tool object or callable.
        """
        if not update and tool in self:
            func_name = getattr(tool, "function_name", tool)
            raise ValueError(f"Tool {func_name} is already registered.")

        if callable(tool):
            tool = func_to_tool(tool)[0]
        if not isinstance(tool, Tool):
            raise TypeError("Please register a Tool object or callable.")

        self.registry[tool.function_name] = tool

    def register_tools(
        self,
        tools: list[REGISTERABLE_TOOL] | REGISTERABLE_TOOL,
    ) -> None:
        """Register multiple tools in the registry.

        Args:
            tools: A single tool or a list of tools to register.
                   Each tool can be a Tool object or a callable.

        Raises:
            ValueError: If any tool is already registered.
            TypeError: If any provided tool is not a Tool object or callable.
        """
        tools_list = tools if isinstance(tools, list) else [tools]
        [
            self.register_tool(tool)
            for tool in to_list(tools_list, dropna=True, flatten=True)
        ]

    @singledispatchmethod
    def match_tool(self, func_call: Any) -> FunctionCalling:
        """Match a function call to a registered tool.

        This method uses single dispatch to handle different input types.

        Args:
            func_call: The function call to match. Can be a tuple, dict,
                       ActionRequest, or string.

        Returns:
            A FunctionCalling object representing the matched tool.

        Raises:
            TypeError: If the input type is not supported.
            ValueError: If function is not registered or call format invalid.
        """
        raise TypeError(f"Unsupported type {type(func_call)}")

    @match_tool.register
    def _(self, func_call: tuple) -> FunctionCalling:
        """Match a function call tuple to a registered tool."""
        if len(func_call) == 2:
            function_name = func_call[0]
            arguments = func_call[1]
            tool = self.registry.get(function_name)
            if not tool:
                raise ValueError(f"Function {function_name} is not registered")
            return FunctionCalling(func_tool=tool, arguments=arguments)
        else:
            raise ValueError(f"Invalid function call {func_call}")

    @match_tool.register
    def _(self, func_call: dict) -> FunctionCalling:
        """Match a function call dictionary to a registered tool."""
        if len(func_call) == 2 and (
            {
                "function",
                "arguments",
            }
            <= func_call.keys()
        ):
            function_name = func_call["function"]
            tool = self.registry.get(function_name)
            if not tool:
                raise ValueError(f"Function {function_name} is not registered")
            return FunctionCalling(
                func_tool=tool,
                arguments=func_call["arguments"],
            )
        raise ValueError(f"Invalid function call {func_call}")

    @match_tool.register
    def _(self, func_call: ActionRequest) -> FunctionCalling:
        """Match an ActionRequest to a registered tool."""
        tool = self.registry.get(func_call.function)
        if not tool:
            func_ = func_call.function
            raise ValueError(f"Function {func_} is not registered.")
        return FunctionCalling(func_tool=tool, arguments=func_call.arguments)

    @match_tool.register
    def _(self, func_call: str) -> FunctionCalling:
        """Parse a string and match it to a registered tool."""
        _call = None
        try:
            _call = to_dict(func_call, str_type="json", fuzzy_parse=True)
        except Exception as e:
            raise ValueError(f"Invalid function call {func_call}") from e

        if isinstance(_call, dict):
            return self.match_tool(_call)
        raise ValueError(f"Invalid function call {func_call}")

    async def invoke(self, func_call: dict | str | ActionRequest) -> Any:
        """Invoke a tool based on the provided function call.

        Args:
            func_call: The function call to invoke. Can be a dictionary,
                       string, or ActionRequest object.

        Returns:
            The result of invoking the matched tool.

        Raises:
            ValueError: If function call can't be matched to registered tool.
        """
        function_calling = self.match_tool(func_call)
        return await function_calling.invoke()

    @property
    def schema_list(self) -> list[dict[str, Any]]:
        """List all tool schemas currently registered in the ToolManager.

        Returns:
            A list of schema dictionaries for all registered tools.
        """
        return [tool.schema_ for tool in self.registry.values()]

    def get_tool_schema(
        self,
        tools: TOOL_TYPE = False,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Retrieve the schema for specific tools or all tools.

        Args:
            tools: Specification of which tools to retrieve schemas for.
                   If True, return all tools. If False, return empty dict.
                   Can also be a specific tool or list of tools.
            **kwargs: Additional keyword arguments to include in output.

        Returns:
            A dictionary containing tool schemas and additional kwargs.

        Raises:
            ValueError: If a specified tool is not registered.
            TypeError: If an unsupported tool type is provided.
        """
        if isinstance(tools, bool):
            if tools:
                tool_kwarg = {"tools": self.schema_list}
            else:
                tool_kwarg = {}
        else:
            tool_kwarg = {"tools": self._get_tool_schema(tools)}
        return tool_kwarg | kwargs

    def _get_tool_schema(
        self,
        tool: Any,
    ) -> dict[str, Any] | list[dict[str, Any]]:
        """Retrieve the schema for a specific tool."""
        if isinstance(tool, dict):
            return tool
        elif isinstance(tool, Tool) or isinstance(tool, str):
            name = tool.function_name if isinstance(tool, Tool) else tool
            if name in self.registry:
                return self.registry[name].schema_
            raise ValueError(f"Tool {name} is not registered.")
        elif isinstance(tool, list):
            return [self._get_tool_schema(t) for t in tool]
        raise TypeError(f"Unsupported type {type(tool)}")


__all__ = ["ToolManager"]
# File: lion_core/action/tool_manager.py



from typing import Any

from lionabc import EventStatus
from lionfuncs import CallDecorator as cd
from lionfuncs import tcall
from pydantic import Field, PrivateAttr
from typing_extensions import override

from lion_core.action.base import ObservableAction
from lion_core.action.tool import Tool
from lion_core.setting import TimedFuncCallConfig


class FunctionCalling(ObservableAction):
    """Represents an action that calls a function with specified arguments.

    Encapsulates a function call, including pre-processing, invocation,
    and post-processing steps. Designed to be executed asynchronously.

    Attributes:
        func_tool (Tool): Tool containing the function to be invoked.
        arguments (dict[str, Any]): Arguments for the function invocation.
        function_name (str | None): Name of the function to be called.
    """

    func_tool: Tool | None = Field(default=None, exclude=True)
    _content_fields: list = PrivateAttr(
        default=["execution_response", "arguments", "function_name"]
    )
    arguments: dict[str, Any] | None = None
    function_name: str | None = None

    def __init__(
        self,
        func_tool: Tool,
        arguments: dict[str, Any],
        timed_config: dict | TimedFuncCallConfig | None = None,
        **kwargs: Any,
    ) -> None:
        """
        kwargs for timed config
        """
        super().__init__(timed_config=timed_config, **kwargs)
        self.func_tool = func_tool
        self.arguments = arguments or {}
        self.function_name = self.func_tool.function_name

    @override
    async def invoke(self) -> Any | None:
        """Asynchronously invokes the function with stored arguments.

        Handles function invocation, applying pre/post-processing steps.
        If a parser is defined, it's applied to the result before returning.

        Returns:
            Any: Result of the function call, possibly processed.

        Raises:
            Exception: If function call or processing steps fail.
        """

        @cd.pre_post_process(
            preprocess=self.func_tool.pre_processor,
            preprocess_kwargs=self.func_tool.pre_processor_kwargs,
        )
        async def _inner(**kwargs) -> tuple[Any, Any | float]:
            kwargs["retry_timing"] = True
            result, elp = await tcall(self.func_tool.function, **kwargs)
            if self.func_tool.post_processor:
                kwargs = self.func_tool.post_processor_kwargs or {}
                kwargs = {
                    **kwargs,
                    **self._timed_config.to_dict(),
                    "retry_timing": True,
                }
                result, elp2 = await tcall(
                    self.func_tool.post_processor, result, **kwargs
                )
                elp += elp2
            return result, elp

        try:
            result, elp = await _inner(**self.arguments)
            self.execution_response = result
            self.execution_time = elp
            self.status = EventStatus.COMPLETED

            if self.func_tool.parser is not None:
                result = self.func_tool.parser(result)

            await self.alog()
            return result

        except Exception as e:
            self.status = EventStatus.FAILED
            self.execution_error = str(e)
            await self.alog()

    def __str__(self) -> str:
        """Returns a string representation of the function call."""
        return f"{self.func_tool.function_name}({self.arguments})"

    def __repr__(self) -> str:
        """Returns a detailed string representation of the function call."""
        return (
            f"FunctionCalling(function={self.func_tool.function_name}, "
            f"arguments={self.arguments})"
        )


__all__ = ["FunctionCalling"]
# File: lion_core/action/function_calling.py



from typing import Any

from lionfuncs import LN_UNDEFINED

from lion_core.communication.message import MessageFlag
from lion_core.communication.system import System


def handle_system(
    system: Any,
    sender: Any | MessageFlag,
    recipient: Any | MessageFlag,
    system_datetime: bool | str | None | MessageFlag,
):
    if isinstance(system, System):
        return system

    if system:
        return System(
            system=system,
            sender=sender,
            recipient=recipient,
            system_datetime=system_datetime,
        )


def validate_system(
    system: Any = None,
    sender=None,
    recipient=None,
    system_datetime=None,
) -> System:
    config = {
        "sender": sender,
        "recipient": recipient,
        "system_datetime": system_datetime,
    }
    config = {k: v for k, v in config.items() if v not in [None, LN_UNDEFINED]}

    if not system:
        return System(DEFAULT_SYSTEM, **config)
    if isinstance(system, System):
        if config:
            for k, v in config.items():
                setattr(system, k, v)
        return system
    return System(system, **config)


DEFAULT_SYSTEM = "You are a helpful AI assistant. Let's think step by step."



import time
from datetime import datetime, timezone

import pytest
from lionabc.exceptions import LionIDError
from pydantic_core._pydantic_core import ValidationError

from lion_core.generic.element import Element
from lion_core.sys_utils import SysUtil


@pytest.fixture
def custom_id():
    return SysUtil.id()


@pytest.fixture
def custom_timestamp():
    return 1625097600.0  # 2021-07-01 00:00:00 UTC


def test_initialization(custom_id, custom_timestamp):
    # Default initialization
    element = Element()
    assert isinstance(element.ln_id, str)
    assert isinstance(element.timestamp, float)

    # Initialization with custom ln_id
    element = Element(ln_id=custom_id)
    assert element.ln_id == custom_id

    # Initialization with custom timestamp
    element = Element(timestamp=custom_timestamp)
    assert element.timestamp == custom_timestamp

    # Initialization with both custom ln_id and timestamp
    element = Element(ln_id=custom_id, timestamp=custom_timestamp)
    assert element.ln_id == custom_id
    assert element.timestamp == custom_timestamp


def test_created_datetime_property(custom_timestamp):
    element = Element(timestamp=custom_timestamp)
    expected_datetime = datetime.fromtimestamp(
        custom_timestamp, tz=timezone.utc
    )
    assert element.created_datetime == expected_datetime
    assert element.created_datetime.tzinfo == timezone.utc


def test_ln_id_validator(custom_id):
    # Valid ln_id format
    element = Element(ln_id=custom_id)
    assert element.ln_id == custom_id

    # Invalid ln_id formats
    invalid_ids = ["", " ", "invalid id", "123", "abc123"]
    for invalid_id in invalid_ids:
        with pytest.raises(LionIDError):
            Element(ln_id=invalid_id)


def test_timestamp_validator():
    # Valid timestamp formats
    valid_timestamps = [
        1625097600.0,  # float
        1625097600,  # int
        "2021-07-01T00:00:00+00:00",  # ISO format string
    ]
    for valid_timestamp in valid_timestamps:
        element = Element(timestamp=valid_timestamp)
        assert isinstance(element.timestamp, float)

    # Invalid timestamp formats
    invalid_timestamps = [
        "invalid_time",
        "2021-07-01",  # Incomplete datetime string
    ]

    with pytest.raises(ValueError):
        Element(timestamp="invalid_time")


def test_from_dict_to_dict(custom_id, custom_timestamp):
    original_dict = {
        "ln_id": custom_id,
        "timestamp": custom_timestamp,
    }
    element = Element.from_dict(original_dict)
    result_dict = element.to_dict()

    assert result_dict["ln_id"] == original_dict["ln_id"]
    assert result_dict["timestamp"] == original_dict["timestamp"]
    assert result_dict["lion_class"] == "Element"

    # Conversion with missing fields
    minimal_dict = {"ln_id": custom_id}
    minimal_element = Element.from_dict(minimal_dict)
    assert minimal_element.ln_id == minimal_dict["ln_id"]
    assert isinstance(minimal_element.timestamp, float)


def test_string_representations(custom_id, custom_timestamp):
    element = Element(ln_id=custom_id, timestamp=custom_timestamp)

    str_output = str(element)
    assert "Element" in str_output
    assert custom_id[:6] in str_output
    assert "2021-07-01" in str_output

    repr_output = repr(element)
    assert "Element" in repr_output
    assert custom_id in repr_output
    assert "1625097600.0" in repr_output


def test_boolean_and_length():
    element = Element()
    assert bool(element) is True
    assert len(element) == 1


def test_element_immutability(custom_id):
    element = Element()
    original_id = element.ln_id
    original_timestamp = element.timestamp

    with pytest.raises(ValidationError):
        element.ln_id = custom_id

    with pytest.raises(ValidationError):
        element.timestamp = time.time()

    assert element.ln_id == original_id
    assert element.timestamp == original_timestamp


def test_element_subclass_behavior():
    class CustomElement(Element):
        custom_attr: str = "default"

    custom_element = CustomElement(custom_attr="custom")
    assert isinstance(custom_element.ln_id, str)
    assert isinstance(custom_element.timestamp, float)
    assert custom_element.custom_attr == "custom"

    assert len(custom_element) == 1
    assert bool(custom_element) is True


def test_element_with_future_timestamp():
    future_timestamp = time.time() + 10000  # 10000 seconds in the future
    element = Element(timestamp=future_timestamp)
    assert element.timestamp > time.time()
    assert element.created_datetime > datetime.now(timezone.utc)


def test_element_serialization_consistency():
    element1 = Element()
    time.sleep(0.1)  # Ensure a different timestamp
    element2 = Element()

    dict1 = element1.to_dict()
    dict2 = element2.to_dict()

    assert dict1["ln_id"] != dict2["ln_id"]
    assert dict1["timestamp"] != dict2["timestamp"]
    assert dict1.keys() == dict2.keys()
    assert dict1["lion_class"] == dict2["lion_class"]


def test_element_large_batch_creation():
    batch_size = 10000
    elements = [Element() for _ in range(batch_size)]

    assert len(elements) == batch_size
    assert (
        len({e.ln_id for e in elements}) == batch_size
    )  # All IDs should be unique

    timestamps = [e.timestamp for e in elements]
    assert (
        max(timestamps) - min(timestamps) < 1.0
    )  # All timestamps should be within 1 second


def test_element_from_dict_with_extra_fields(custom_id):
    extra_dict = {
        "ln_id": custom_id,
        "timestamp": time.time(),
    }
    element = Element.from_dict(extra_dict)

    assert element.ln_id == extra_dict["ln_id"]
    assert element.timestamp == extra_dict["timestamp"]


def test_element_to_dict_consistency():
    element = Element()
    dict_repr = element.to_dict()

    assert dict_repr["ln_id"] == element.ln_id
    assert dict_repr["timestamp"] == element.timestamp
    assert dict_repr["lion_class"] == element.__class__.__name__

    # Ensure no extra fields are present
    assert set(dict_repr.keys()) == {"ln_id", "timestamp", "lion_class"}


# File: tests/test_generic/test_element.py



from typing import Any, Literal

from lion_core.communication.instruction import Instruction
from lion_core.communication.message import MessageFlag


def handle_instruction(
    sender: Any | MessageFlag,
    recipient: Any | MessageFlag,
    instruction: Any | MessageFlag,
    context: Any | MessageFlag,
    guidance: Any | MessageFlag,
    request_fields: dict | MessageFlag,
    images: list | MessageFlag,
    image_detail: Literal["low", "high", "auto"] | MessageFlag,
):
    if isinstance(instruction, Instruction):
        if context:
            instruction.update_context(context)
        if request_fields:
            instruction.update_request_fields(
                request_fields=request_fields,
            )
        if images:
            instruction.update_images(
                images=images,
                image_detail=image_detail,
            )
        if guidance:
            instruction.update_guidance(
                guidance=guidance,
            )
        return instruction

    return Instruction(
        instruction=instruction,
        context=context,
        guidance=guidance,
        images=images,
        sender=sender,
        recipient=recipient,
        request_fields=request_fields,
        image_detail=image_detail,
    )



import contextlib
from collections.abc import Iterator
from typing import Any

from lionabc import Collective
from lionabc.exceptions import ItemNotFoundError, LionTypeError, LionValueError
from lionfuncs import LN_UNDEFINED
from pydantic import Field
from pydantic_core import PydanticUndefined
from typing_extensions import Self

from lion_core.generic.element import Element
from lion_core.generic.pile import Pile, pile
from lion_core.generic.progression import Progression, prog


class Flow(Element):
    progressions: Pile[Progression] = Field(
        default_factory=lambda: pile({}, item_type={Progression})
    )

    registry: dict = Field(default_factory=dict)
    default_name: str = Field(None)

    def __init__(
        self,
        progressions: Pile[Progression] | None = None,
        default_name: str | None = None,
    ) -> None:
        """
        Initializes a Flow instance.

        Args:
            sequences (optional): Initial sequences to include in the flow.
            default_name (optional): Default name for the flow.
        """
        super().__init__()
        self.progressions = self._validate_progressions(progressions)

        if default_name:
            self.default_name = default_name

        for prog_ in self.progressions:
            a = prog_.name or prog_.ln_id
            self.register(prog_, a)

    def _validate_progressions(self, value: Any) -> Pile[Progression]:
        try:
            if isinstance(value, Collective):
                if Progression in getattr(value, "item_type", []):
                    return value
                else:
                    value = list(value)
            return pile(value, item_type={Progression})

        except Exception as e:
            raise LionValueError(f"Invalid progressions: {e}")

    def all_orders(self) -> list[list[str]]:
        """Get all progression orders as a list of lists."""
        return [list(seq) for seq in self.progressions]

    def unique(self) -> list[str]:
        """Get a list of unique items across all progressions."""
        return list({item for seq in self.progressions for item in seq})

    def keys(self) -> Iterator[str]:
        """Yield the keys (IDs) of all progressions."""
        yield from self.progressions.keys()

    def values(self) -> Iterator[Progression]:
        """Yield the values (Progression objects) of all progressions."""
        yield from self.progressions.values()

    def items(self) -> Iterator[tuple[str, Progression]]:
        """Yield the (key, value) pairs of all progressions."""
        yield from self.progressions.items()

    def __getitem__(
        self,
        prog_: str | Progression = None,
        default: Any = LN_UNDEFINED,
    ) -> Progression | Any:
        """Get a progression by its ID or name."""
        return self.get(prog_, default)

    def __setitem__(
        self,
        prog_: Progression | str,
        index: int | slice | None = None,
        value: Any = None,
        /,
    ) -> None:
        """Set a progression or an item within a progression."""
        if prog_ not in self:
            raise ItemNotFoundError(f"Sequence {prog_}")

        if index:
            self.progressions[prog_][index] = value
            return

        self.progressions[prog_] = value

    def __contains__(self, item: Any) -> bool:
        """Check if an item is in any progression or in the registry."""
        check = item in self.registry
        if check is False:
            check = item in self.progressions
        if check is False:
            check = item in self.unique()
        return check

    def shape(self) -> tuple[int, list[int]]:
        """
        Get the shape of the Flow.

        Returns:
            A tuple containing the number of progressions and a list of
            their lengths.
        """
        ao = self.all_orders()
        return (len(ao), [len(i) for i in ao])

    def size(self) -> int:
        """
        Get the total number of items across all progressions.

        Returns:
            The total number of items in the Flow.
        """

        return sum(len(seq) for seq in self.all_orders())

    def clear(self) -> None:
        """Clear all progressions and the registry."""
        self.progressions.clear()
        self.registry.clear()

    def include(
        self,
        prog_: str | Progression | None = None,
        /,
        item: Any = None,
        name: str | None = None,
    ) -> bool | None:
        """
        Include a progression or an item in a progression.

        Args:
            prog_: The progression to include or modify.
            item: The item to include in the progression.
            name: The name for the progression.

        Returns:
            True if the inclusion was successful, False otherwise.
        """
        _sequence = self._find_prog(prog_, None)
        if _sequence is None:
            _sequence = self._find_prog(name, None)

        if not _sequence:
            if not item and not name:
                """None is not in the registry or sequencees."""
                return False
            if item:
                self.append(item, name)
                return item in self

        else:
            if _sequence in self:
                if not item and not name:
                    return True
                if item:
                    self.append(item, _sequence)
                    return item in self
                return True  # will ignore name if sequence is already found

            else:
                if isinstance(prog_, Progression):
                    if item and prog_.include(item):
                        self.register(prog_, name)
                    return prog_ in self

                return False

    def exclude(
        self,
        prog_: Progression | str | None = None,
        /,
        item: Any = None,
        name: str | None = None,
    ) -> bool | Any | None:
        """
        Exclude a progression or an item from a progression.

        Args:
            seq: The progression to exclude or modify.
            item: The item to exclude from the progression.
            name: The name of the progression to exclude.

        Returns:
            True if the exclusion was successful, False otherwise.
        """
        # if sequence is not None, we will not check the name
        if prog_ is not None:
            with contextlib.suppress(ItemNotFoundError, AttributeError):
                if item:
                    # if there is item, we exclude it from the sequence
                    self.progressions[self.registry[prog_]].exclude(item)
                    return item not in self.progressions[self.registry[prog_]]
                else:
                    # if there is no item, we exclude the sequence
                    a = self.registry.pop(prog_.name or prog_.ln_id, None)
                    return a is not None and self.progressions.exclude(prog_)
            return False

        if name is not None:
            with contextlib.suppress(ItemNotFoundError):
                if item:
                    # if there is item, we exclude it from the sequence
                    return self.progressions[self.registry[name]].exclude(item)
                else:
                    # if there is no item, we exclude the sequence
                    a = self.registry.pop(name, None)
                    return a is not None and self.progressions.exclude(a)
            return False

    def register(self, prog_: Progression, /, name: str | None = None) -> None:
        """
        Register a new progression.

        Args:
            prog_: The Progression object to register.
            name: The name for the progression.

        Raises:
            LionTypeError: If prog_ is not a Progression.
            ValueError: If the name already exists in the registry.
        """
        if not isinstance(prog_, Progression):
            raise LionTypeError("Sequence must be of type Progression.")

        name = name or prog_.name
        if not name:
            if self.default_name in self.registry:
                name = prog_.ln_id
            else:
                name = self.default_name

        if name in self.registry:
            raise ValueError(f"Sequence '{name}' already exists.")

        self.progressions.include(prog_)
        self.registry[name] = prog_.ln_id

    def append(
        self, item: Any, prog_: str | Progression | None = None, /
    ) -> None:
        """
        Append an item to a progression.

        Args:
            item: The item to append.
            prog_: The progression to append to (optional).
        """
        if not prog_:
            if self.default_name in self.registry:
                prog_ = self.registry[self.default_name]
                self.progressions[prog_].include(item)
                return

            p = prog(item, self.default_name)
            self.register(p)
            return

        if prog_ in self.progressions:
            self.progressions[prog_].include(item)
            return

        if prog_ in self.registry:
            self.progressions[self.registry[prog_]].include(item)
            return

        p = prog(item, prog_ if isinstance(prog_, str) else None)
        self.register(p)

    def popleft(self, prog_: str | Progression | None = None, /) -> str | Any:
        """
        Remove and return the leftmost item from a progression.

        Args:
            prog_: The progression to pop from (optional).

        Returns:
            The leftmost item from the specified progression.
        """
        prog_ = self._find_prog(prog_)
        return self.progressions[prog_].popleft()

    def get(
        self,
        prog_: Progression | str | None = None,
        /,
        default: Any = LN_UNDEFINED,
    ) -> Progression | Any:
        """
        Get a progression by its ID or name.

        Args:
            prog_: The progression ID or name.
            default: Default value if progression is not found.

        Returns:
            The requested Progression or the default value.

        Raises:
            ItemNotFoundError: If no progression is found and no default
                is provided.
        """
        prog_ = getattr(prog_, "ln_id", None) or prog_

        if prog_ is None:
            if self.default_name in self.registry:
                return self.progressions[self.registry[self.default_name]]
            if default in [LN_UNDEFINED, PydanticUndefined]:
                raise ItemNotFoundError("No progression found.")

        if prog_ in self.registry:
            return self.progressions[self.registry[prog_]]

        try:
            return self.progressions[prog_]
        except KeyError as e:
            if default in [LN_UNDEFINED, PydanticUndefined]:
                raise e
            return default

    def remove(
        self, item: Any, prog_: str | Progression | None = None, /
    ) -> None:
        """
        Remove an item from one or all progressions.

        Args:
            item: The item to remove.
            prog_: The progression to remove from, or "all" for all
                    progressions.
        """
        if prog_ == "all":
            for seq in self.progressions:
                seq.remove(item)
            return

        prog_ = self._find_prog(prog_)
        self.progressions[prog_].remove(item)

    def __len__(self) -> int:
        return len(self.progressions)

    def __iter__(self) -> Iterator[Progression]:
        return iter(self.progressions)

    def __next__(self) -> Progression:
        return next(self.__iter__())

    def _find_prog(
        self,
        prog_: str | Progression | None = None,
        default: Any = LN_UNDEFINED,
    ):
        if not prog_:
            if self.default_name in self.registry:
                return self.registry[self.default_name]
            if default not in [LN_UNDEFINED, PydanticUndefined]:
                return default
            raise ItemNotFoundError("No progression found.")

        if prog_ in self.progressions:
            return prog_.ln_id if isinstance(prog_, Progression) else prog_

        if prog_ in self.registry:
            return self.registry[prog_]

    def to_dict(self) -> dict[str, Any]:
        return {
            "progressions": self.progressions.to_dict(),
            "default_name": self.default_name,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        progressions = Pile.from_dict(data["progressions"])
        return cls(progressions, data["default_name"])


def flow(
    progressions: Pile[Progression] | None = None,
    default_name: str | None = None,
):
    return Flow(progressions, default_name)


__all__ = ["Flow", "flow"]

# Flow: lion_core/generic/flow.py



from .action_msg import handle_action, handle_action_response
from .assistant_msg import handle_assistant
from .create_msg import create_message
from .create_request import create_action_request
from .extract_request import extract_action_request
from .instruction_msg import handle_instruction
from .system_msg import handle_system
from .validate_msg import validate_message

__all__ = [
    "handle_action",
    "handle_action_response",
    "handle_assistant",
    "create_message",
    "create_action_request",
    "extract_action_request",
    "handle_instruction",
    "handle_system",
    "validate_message",
]



import pytest

from lion_core.communication.instruction import (
    Instruction,
    format_image_content,
    prepare_instruction_content,
    prepare_request_response_format,
)
from lion_core.communication.message import MessageFlag, MessageRole
from lion_core.form.form import Form
from lion_core.generic.note import Note
from lion_core.sys_utils import SysUtil


# Helper function to create a mock BaseTask
class MockBaseTask(Form):
    def __init__(self, instruction_dict):
        super().__init__()
        self.instruction_dict = instruction_dict
        self.ln_id = "mock_task_id"


# Tests for helper functions
def test_prepare_request_response_format():
    fields = {"field1": "description1", "field2": "description2"}
    result = prepare_request_response_format(fields)
    assert "```json" in result
    assert "field1" in result
    assert "field2" in result


def test_format_image_content():
    text = "Sample text"
    images = ["base64_image1", "base64_image2"]
    result = format_image_content(text, images, "low")
    assert len(result) == 3
    assert result[0]["type"] == "text"
    assert result[1]["type"] == "image_url"
    assert result[2]["type"] == "image_url"


def test_prepare_instruction_content():
    result = prepare_instruction_content(
        instruction="Test instruction",
        context="Test context",
        request_fields={"field1": "desc1"},
        images=["image1"],
        image_detail="high",
    )
    assert isinstance(result, Note)
    assert result.get("instruction") == "Test instruction"
    assert result.get("context") == ["Test context"]
    assert "request_response_format" in result.to_dict()


# Tests for Instruction class
def test_instruction_init():
    instruction = Instruction("Test instruction", context="Test context")
    assert instruction.role == MessageRole.USER
    assert instruction.sender == "user"
    assert instruction.recipient == "N/A"
    assert instruction.instruction == "Test instruction"


def test_instruction_init_with_all_params():
    instruction = Instruction(
        "Test instruction",
        context="Test context",
        images=["image1"],
        sender=SysUtil.id(),
        recipient=SysUtil.id(),
        request_fields={"field1": "desc1"},
        image_detail="high",
    )
    assert "images" in instruction.content.to_dict()
    assert instruction.content.get("image_detail") == "high"


def test_instruction_init_with_message_load():
    protected_params = {
        "role": MessageRole.USER,
        "content": Note(instruction="Test"),
        "sender": "user",
    }
    instruction = Instruction(
        MessageFlag.MESSAGE_LOAD,
        context=MessageFlag.MESSAGE_LOAD,
        images=MessageFlag.MESSAGE_LOAD,
        sender=MessageFlag.MESSAGE_LOAD,
        recipient=MessageFlag.MESSAGE_LOAD,
        request_fields=MessageFlag.MESSAGE_LOAD,
        image_detail=MessageFlag.MESSAGE_LOAD,
        guidance=MessageFlag.MESSAGE_LOAD,
        protected_init_params=protected_params,
    )
    assert instruction.role == MessageRole.USER
    assert instruction.instruction == "Test"


def test_instruction_init_with_message_clone():
    instruction = Instruction(
        MessageFlag.MESSAGE_CLONE,
        context=MessageFlag.MESSAGE_CLONE,
        images=MessageFlag.MESSAGE_CLONE,
        sender=MessageFlag.MESSAGE_CLONE,
        recipient=MessageFlag.MESSAGE_CLONE,
        request_fields=MessageFlag.MESSAGE_CLONE,
        guidance=MessageFlag.MESSAGE_LOAD,
        image_detail=MessageFlag.MESSAGE_CLONE,
    )
    assert instruction.role == MessageRole.USER


def test_instruction_update_request_fields():
    instruction = Instruction("Test", request_fields={"field1": "desc1"})
    instruction.update_request_fields({"field2": "desc2"})
    assert "field1" in instruction.content["request_fields"]
    assert "field2" in instruction.content["request_fields"]


def test_instruction_update_context():
    instruction = Instruction("Test")
    instruction.update_context("context1", "context2", key="value")
    assert len(instruction.content["context"]) == 3
    assert "context1" in instruction.content["context"]
    assert {"key": "value"} in instruction.content["context"]


def test_instruction_format_content():
    instruction = Instruction("Test", images=["image1"], image_detail="low")
    formatted = instruction._format_content()
    assert isinstance(formatted["content"], list)
    assert formatted["content"][0]["type"] == "text"
    assert formatted["content"][1]["type"] == "image_url"


# Edge cases and additional tests
def test_instruction_empty_init():
    instruction = Instruction(None)
    assert instruction.instruction == "N/A"


def test_instruction_with_large_context():
    large_context = ["context"] * 1000
    instruction = Instruction("Test", context=large_context)
    assert len(instruction.content["context"]) == 1000


def test_instruction_with_many_images():
    many_images = ["image"] * 100
    instruction = Instruction("Test", images=many_images)
    assert len(instruction.content["images"]) == 100


def test_instruction_update_context_empty():
    instruction = Instruction("Test")
    instruction.update_context()
    assert "context" in instruction.content.to_dict()
    assert instruction.content["context"] == []


def test_instruction_format_content_no_images():
    instruction = Instruction("Test")
    formatted = instruction._format_content()
    assert isinstance(formatted["content"], str)


def test_instruction_with_unicode():
    instruction = Instruction("Test 你好", context="Context こんにちは")
    assert "你好" in instruction.instruction
    assert "こんにちは" in instruction.content["context"][0]


def test_instruction_with_special_characters():
    special_chars = "!@#$%^&*()_+{}[]|\\:;\"'<>,.?/~`"
    instruction = Instruction(special_chars)
    assert instruction.instruction == special_chars


def test_instruction_with_very_long_instruction():
    long_instruction = "a" * 10000
    instruction = Instruction(long_instruction)
    assert len(instruction.instruction) == 10000


def test_instruction_update_request_fields_multiple_times():
    instruction = Instruction("Test", request_fields={"field1": "desc1"})
    instruction.update_request_fields({"field2": "desc2"})
    instruction.update_request_fields({"field3": "desc3"})
    assert len(instruction.content["request_fields"]) == 3


def test_instruction_update_context_multiple_times():
    instruction = Instruction("Test")
    instruction.update_context("context1")
    instruction.update_context("context2")
    instruction.update_context(key="value")
    assert len(instruction.content["context"]) == 3


def test_instruction_with_empty_request_fields():
    instruction = Instruction("Test", request_fields={})
    assert "instruction" in instruction.content.to_dict()
    assert instruction.content["instruction"] == "Test"


def test_instruction_with_none_values():
    instruction = Instruction(
        "Test", context=None, images=None, request_fields=None
    )
    assert "context" not in instruction.content.to_dict()
    assert "images" not in instruction.content.to_dict()
    assert "request_fields" not in instruction.content.to_dict()


def test_instruction_update_request_fields_override():
    instruction = Instruction("Test", request_fields={"field1": "desc1"})
    instruction.update_request_fields({"field1": "new_desc"})
    assert instruction.content["request_fields"]["field1"] == "new_desc"


# Performance test
def test_instruction_performance():
    import time

    start_time = time.time()
    for _ in range(1000):
        Instruction(
            "Test",
            context="Context",
            images=["image1"],
            request_fields={"field1": "desc1"},
        )
    end_time = time.time()
    assert (
        end_time - start_time < 1
    )  # Should create 1000 instructions in less than 1 second


# Thread safety test
def test_instruction_thread_safety():
    import threading

    def create_instruction():
        Instruction("Test", context=f"Context {threading.get_ident()}")

    threads = [threading.Thread(target=create_instruction) for _ in range(100)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()


# Test with maximum possible fields
def test_instruction_max_fields():
    max_fields = {f"field{i}": f"desc{i}" for i in range(1000)}
    instruction = Instruction("Test", request_fields=max_fields)
    assert len(instruction.content["request_fields"]) == 1000


# Test with very large base64 image
def test_instruction_large_image():
    large_image = "a" * 1000000  # 1MB base64 string
    instruction = Instruction("Test", images=[large_image])
    assert len(instruction.content["images"][0]) == 1000000


# Test instruction with all possible image details
@pytest.mark.parametrize("image_detail", ["low", "high", "auto"])
def test_instruction_all_image_details(image_detail):
    instruction = Instruction(
        "Test", images=["image1"], image_detail=image_detail
    )
    assert instruction.content["image_detail"] == image_detail


# Test instruction with mixed types in context
def test_instruction_mixed_context_types():
    mixed_context = ["string", 123, {"key": "value"}, ["list", "items"]]
    instruction = Instruction("Test", context=mixed_context)
    assert instruction.content["context"] == mixed_context


# Test updating context with various types
def test_instruction_update_context_various_types():
    instruction = Instruction("Test")
    instruction.update_context(
        "string", 123, {"key": "value"}, ["list", "items"]
    )
    assert len(instruction.content["context"]) == 4
    assert isinstance(instruction.content["context"][1], int)
    assert isinstance(instruction.content["context"][2], dict)


# Test clone method
def test_instruction_clone():
    original = Instruction("Test", context="Original context")
    cloned = original.clone()
    assert original.instruction == cloned.instruction


# Test for potential memory leaks
def test_instruction_memory_usage():
    import sys

    initial_size = sys.getsizeof(Instruction("Test"))
    large_instruction = Instruction("Test", context="a" * 1000000)
    large_size = sys.getsizeof(large_instruction)
    assert (
        large_size - initial_size < 2000000
    )  # Ensure no significant unexpected memory increase


# Test with invalid UTF-8 sequences
def test_instruction_invalid_utf8():
    invalid_utf8 = b"Invalid UTF-8: \xff"
    with pytest.raises(UnicodeDecodeError):
        Instruction(invalid_utf8.decode("utf-8"))



from typing import ClassVar

from lionabc import AbstractSpace, BaseiModel
from pydantic import Field, model_validator

from lion_core.communication.system import System
from lion_core.generic.node import Node
from lion_core.generic.pile import Pile
from lion_core.session.msg_handlers.system_msg import validate_system
from lion_core.sys_utils import SysUtil


class BaseSession(Node, AbstractSpace):
    system: System | None = Field(None)
    user: str | None = Field(None)
    imodel: BaseiModel | None = Field(None)
    name: str | None = Field(None)
    pile_type: ClassVar[type] = Pile

    @model_validator(mode="before")
    def validate_system(cls, data: dict):
        system = data.pop("system", None)
        sender = data.pop("system_sender", None)
        system_datetime = data.pop("system_datetime", None)

        system = validate_system(
            system=system,
            sender=sender,
            system_datetime=system_datetime,
        )
        data["system"] = system
        return data

    @model_validator(mode="after")
    def check_system_recipient(self):
        if not SysUtil.is_id(self.system.recipient):
            self.system.recipient = self.ln_id
        return self


__all__ = ["BaseSession"]
# File: lion_core/session/base.py



import asyncio
import json
from typing import Any

import pytest

from lion_core.action.tool import Tool, func_to_tool


# Sample functions for testing
def sample_function(x: int, y: int) -> int:
    """
    A sample function that adds two numbers.

    Args:
        x: The first number.
        y: The second number.

    Returns:
        The sum of x and y.
    """
    return x + y


async def async_sample_function(x: int, y: int) -> int:
    """
    An async sample function that adds two numbers.

    Args:
        x: The first number.
        y: The second number.

    Returns:
        The sum of x and y.
    """
    return x + y


def sample_preprocessor(args: dict[str, Any]) -> dict[str, Any]:
    """A sample preprocessor that doubles the input values."""
    return {k: v * 2 for k, v in args.items()}


def sample_postprocessor(result: int) -> str:
    """A sample postprocessor that converts the result to a string."""
    return f"Result: {result}"


def sample_parser(result: Any) -> dict[str, Any]:
    """A sample parser that wraps the result in a dictionary."""
    return {"parsed_result": result}


# Test Tool class
def test_tool_init():
    tool = Tool(function=sample_function)
    assert callable(tool.function)
    assert tool.schema_ is not None
    assert tool.schema_["function"]["name"] == "sample_function"


def test_tool_init_with_schema():
    custom_schema = {"function": {"name": "custom_function"}}
    tool = Tool(function=sample_function, schema_=custom_schema)
    assert tool.schema_ == custom_schema


def test_tool_function_name():
    tool = Tool(function=sample_function)
    assert tool.function_name == "sample_function"


def test_tool_str_representation():
    tool = Tool(function=sample_function)
    str_repr = str(tool)
    assert "Tool(ln_id=" in str_repr
    assert "timestamp=" in str_repr
    assert "schema_=" in str_repr


def test_tool_serialization():
    tool = Tool(
        function=sample_function,
        pre_processor=sample_preprocessor,
        post_processor=sample_postprocessor,
        parser=sample_parser,
        pre_processor_kwargs={"arg1": "value1"},
        post_processor_kwargs={"arg2": "value2"},
    )
    serialized = tool.model_dump_json()
    deserialized = json.loads(serialized)

    assert deserialized["function"] == "sample_function"
    assert deserialized["pre_processor"] == "sample_preprocessor"
    assert deserialized["post_processor"] == "sample_postprocessor"
    assert deserialized["parser"] == "sample_parser"
    assert json.loads(deserialized["pre_processor_kwargs"]) == {
        "arg1": "value1"
    }
    assert json.loads(deserialized["post_processor_kwargs"]) == {
        "arg2": "value2"
    }


# Test func_to_tool function
def test_func_to_tool_single_function():
    tools = func_to_tool(sample_function)
    assert len(tools) == 1
    assert isinstance(tools[0], Tool)
    assert tools[0].function == sample_function


def test_func_to_tool_multiple_functions():
    functions = [sample_function, async_sample_function]
    tools = func_to_tool(functions)
    assert len(tools) == 2
    assert all(isinstance(tool, Tool) for tool in tools)
    assert [tool.function for tool in tools] == functions


def test_func_to_tool_with_parser():
    tools = func_to_tool(sample_function, parser=sample_parser)
    assert len(tools) == 1
    assert tools[0].parser == sample_parser


def test_func_to_tool_multiple_functions_with_parsers():
    functions = [sample_function, async_sample_function]
    parsers = [sample_parser, None]
    tools = func_to_tool(functions, parser=parsers)
    assert len(tools) == 2
    assert tools[0].parser == sample_parser
    assert tools[1].parser is None


def test_func_to_tool_docstring_style():
    tools = func_to_tool(sample_function, docstring_style="rest")
    assert (
        tools[0].schema_["function"]["description"]
        == "A sample function that adds two numbers."
    )


def test_func_to_tool_additional_kwargs():
    tools = func_to_tool(sample_function, pre_processor=sample_preprocessor)
    assert tools[0].pre_processor == sample_preprocessor


def test_func_to_tool_mismatched_parsers():
    with pytest.raises(ValueError):
        func_to_tool(
            [sample_function, async_sample_function], parser=[sample_parser]
        )


# Edge cases and additional tests
def test_tool_with_lambda_function():
    lambda_func = lambda x, y: x * y
    tool = Tool(function=lambda_func)
    assert tool.function == lambda_func
    assert "lambda" in tool.function_name.lower()


def test_tool_with_class_method():
    class SampleClass:
        @classmethod
        def class_method(cls, x: int) -> int:
            return x * 2

    tool = Tool(function=SampleClass.class_method)
    assert tool.function == SampleClass.class_method
    assert tool.function_name == "class_method"


def test_tool_with_static_method():
    class SampleClass:
        @staticmethod
        def static_method(x: int) -> int:
            return x * 3

    tool = Tool(function=SampleClass.static_method)
    assert tool.function == SampleClass.static_method
    assert tool.function_name == "static_method"


def test_func_to_tool_empty_input():
    func_to_tool([])


def test_func_to_tool_non_callable():
    with pytest.raises(AttributeError):
        func_to_tool("not_a_function")


def test_tool_with_function_without_docstring():
    def no_docstring_func(x: int) -> int:
        return x + 1

    tool = Tool(function=no_docstring_func)
    assert tool.schema_["function"]["description"] == None


def test_func_to_tool_preserve_metadata():
    def metadata_func() -> None:
        """Test function with metadata."""
        pass

    metadata_func.custom_attr = "test_value"

    tools = func_to_tool(metadata_func)
    assert hasattr(tools[0].function, "custom_attr")
    assert tools[0].function.custom_attr == "test_value"


@pytest.mark.asyncio
async def test_tool_with_async_function():
    tool = Tool(function=async_sample_function)
    assert asyncio.iscoroutinefunction(tool.function)


def test_tool_schema_generation():
    def typed_func(a: int, b: str, c: bool = False) -> dict:
        """
        A function with typed parameters.

        Args:
            a: An integer parameter
            b: A string parameter
            c: A boolean parameter with default value

        Returns:
            A dictionary
        """
        return {"a": a, "b": b, "c": c}

    tool = Tool(function=typed_func)
    schema = tool.schema_["function"]
    assert schema["name"] == "typed_func"
    assert (
        "An integer parameter"
        in schema["parameters"]["properties"]["a"]["description"]
    )
    assert schema["parameters"]["properties"]["a"]["type"] == "number"
    assert schema["parameters"]["properties"]["b"]["type"] == "string"
    assert schema["parameters"]["properties"]["c"]["type"] == "boolean"
    assert "c" in schema["parameters"]["required"]


def test_tool_with_variable_arguments():
    def var_args_func(*args, **kwargs):
        pass

    tool = Tool(function=var_args_func)
    assert "parameters" in tool.schema_["function"]



from typing import Any

from lionabc import BaseiModel
from lionabc.exceptions import ItemNotFoundError, LionValueError
from lionfuncs import LN_UNDEFINED
from pydantic import Field, PrivateAttr

from lion_core.action.tool_manager import ToolManager
from lion_core.communication.mail_manager import MailManager
from lion_core.communication.message import RoledMessage
from lion_core.generic.exchange import Exchange
from lion_core.generic.flow import Flow
from lion_core.generic.pile import Pile, pile
from lion_core.generic.progression import Progression, prog
from lion_core.generic.utils import to_list_type
from lion_core.session.base import BaseSession
from lion_core.session.branch import Branch
from lion_core.sys_utils import SysUtil


class Session(BaseSession):
    """
    Manages multiple conversation branches and mail transfer in a session.

    Attributes:
        branches (Pile | None): Collection of conversation branches.
        default_branch (Branch | None): The default conversation branch.
        mail_transfer (Exchange | None): Mail transfer system.
        mail_manager (MailManager | None): Manages mail operations.
        conversations (Flow | None): Manages conversation flow.
    """

    branches: Pile | None = Field(None)
    default_branch: Branch | None = Field(None, exclude=True)
    mail_transfer: Exchange | None = Field(None)
    mail_manager: MailManager | None = Field(None, exclude=True)
    conversations: Flow | None = Field(None)
    _branch_type: type[Branch] = PrivateAttr(Branch)

    async def new_branch(
        self,
        system: Any = None,
        system_sender: str | None = None,
        system_datetime: Any = None,
        user: str | None = None,
        name: str | None = None,
        imodel: BaseiModel | None = None,
        messages: Pile | None = None,
        progress: Progression | None = None,
        tool_manager: ToolManager | None = None,
        tools: Any = None,
        **kwargs,  # additional branch parameters
    ) -> Branch:
        if system in [None, LN_UNDEFINED]:
            system = self.system.clone()
            system.sender = self.ln_id
            system_sender = self.ln_id

        branch = self._branch_type(
            system=system,
            system_sender=system_sender,
            system_datetime=system_datetime,
            name=name,
            user=user,
            imodel=imodel or self.imodel,
            messages=messages,
            progress=progress,
            tool_manager=tool_manager,
            tools=tools,
            **kwargs,
        )

        self.conversations.register(branch.progress, name=name)
        self.branches.include(branch)
        self.mail_manager.add_sources(branch)
        if self.default_branch is None:
            self.default_branch = branch
        return branch

    def remove_branch(
        self,
        branch: Branch | str,
        delete: bool = False,
    ):
        branch = SysUtil.get_id(branch)

        if branch not in self.branches:
            _s = (
                str(branch)
                if len(str(branch)) < 10
                else str(branch)[:10] + "..."
            )
            raise ItemNotFoundError(f"Branch {_s}.. does not exist.")
        branch: Branch = self.branches[branch]

        self.conversations.exclude(prog=branch.progress)
        self.branches.exclude(branch)
        self.mail_manager.delete_source(branch.ln_id)

        if self.default_branch.ln_id == branch.ln_id:
            if self.branches.is_empty():
                self.default_branch = None
            else:
                self.default_branch = self.branches[0]

        if delete:
            del branch

    def split_branch(self, branch: Branch | str) -> Branch:
        """
        Split a branch, creating a new branch with the same messages and tools.

        Args:
            branch: The branch to split or its identifier.

        Returns:
            The newly created branch.
        """
        branch: Branch = self.branches[branch]
        system = branch.system.clone() if branch.system else None
        if system:
            system.sender = branch.ln_id
        progress = prog()
        messages = pile({}, RoledMessage, strict=False)

        for id_ in branch.progress:
            clone_message: RoledMessage = branch.messages[id_].clone()
            progress.append(clone_message.ln_id)
            messages.append(clone_message)

        tools = (
            list(branch.tool_manager.registry.values())
            if branch.tool_manager.registry
            else None
        )
        branch_clone = self._branch_type(
            system=system,
            system_sender=branch.ln_id,
            user=branch.user,
            progress=progress,
            messages=messages,
            tools=tools,
        )
        for message in branch_clone.messages:
            message.sender = branch.ln_id
            message.recipient = branch_clone.ln_id
        self.branches.append(branch_clone)
        self.mail_manager.add_sources(branch_clone)
        return branch_clone

    def change_default_branch(self, branch: Branch | str):
        """
        Change the default branch of the session.

        Args:
            branch: The branch to set as default or its identifier.
        """
        branch = self.branches[branch]
        if branch and len(branch) == 1:
            self.default_branch = branch
        raise LionValueError("Session can only have one default branch.")

    def collect(self, from_: Branch | str | Pile[Branch] | None = None):
        """
        Collect mail from specified branches.

        Args:
            from_: The branches to collect mail from. If None, collect
                from all.

        Raises:
            ValueError: If mail collection fails.
        """
        if from_ is None:
            self.mail_manager.collect_all()
        else:
            try:
                sources = to_list_type(from_)
                for source in sources:
                    self.mail_manager.collect(SysUtil.get_id(source))
            except Exception as e:
                raise ValueError(f"Failed to collect mail. Error: {e}")

    def send(self, to_: Branch | str | Pile[Branch] | None = None):
        """
        Send mail to specified branches.

        Args:
            to_: The branches to send mail to. If None, send to all.

        Raises:
            ValueError: If mail sending fails.
        """
        if to_ is None:
            self.mail_manager.send_all()
        else:
            try:
                sources = to_list_type(to_)
                for source in sources:
                    self.mail_manager.send(SysUtil.get_id(source))
            except Exception as e:
                raise ValueError(f"Failed to send mail. Error: {e}")

    def collect_send_all(self, receive_all: bool = False):
        """
        Collect and send mail for all branches, optionally receiving all mail.

        Args:
            receive_all: If True, receive all mail for all branches.
        """
        self.collect()
        self.send()
        if receive_all:
            for branch in self.branches:
                branch.receive_all()


__all__ = ["Session"]
# File: lion_core/session/session.py



import asyncio
import random
import string
from typing import Any

import pytest
from lionabc.exceptions import ItemNotFoundError
from pydantic import Field

from lion_core.generic.element import Element
from lion_core.generic.progression import Progression
from lion_core.sys_utils import SysUtil


class MockElement(Element):
    value: Any = Field(None)


@pytest.fixture
def sample_elements():
    return [MockElement(value=i) for i in range(5)]


@pytest.fixture
def sample_progression(sample_elements):
    return Progression(order=[e.ln_id for e in sample_elements])


def generate_random_string(length: int) -> str:
    return "".join(
        random.choices(string.ascii_letters + string.digits, k=length)
    )


@pytest.mark.parametrize(
    "input_data",
    [
        [],
        [MockElement(value=i) for i in range(3)],
    ],
)
def test_initialization(input_data):
    prog = Progression(order=input_data)
    assert len(prog) == len(input_data)
    for item in input_data:
        assert SysUtil.get_id(item) in prog.order


def test_initialization_with_name():
    name = "test_progression"
    prog = Progression(name=name)
    assert prog.name == name


def test_contains(sample_progression, sample_elements):
    for element in sample_elements:
        assert element in sample_progression
        assert element.ln_id in sample_progression


def test_not_contains(sample_progression):
    assert "non_existent_id" not in sample_progression
    assert MockElement(value="new") not in sample_progression


def test_list_conversion(sample_progression, sample_elements):
    prog_list = sample_progression.__list__()
    assert isinstance(prog_list, list)
    assert len(prog_list) == len(sample_elements)
    assert all(isinstance(item, str) for item in prog_list)


def test_len(sample_progression, sample_elements):
    assert len(sample_progression) == len(sample_elements)


@pytest.mark.parametrize(
    "index, expected_type",
    [
        (0, str),
        (slice(0, 2), Progression),
    ],
)
def test_getitem(sample_progression, index, expected_type):
    result = sample_progression[index]
    assert isinstance(result, expected_type)


def test_getitem_out_of_range(sample_progression):
    with pytest.raises(ItemNotFoundError):
        _ = sample_progression[len(sample_progression)]


def test_setitem(sample_progression):
    new_element = MockElement(value="new")
    sample_progression[0] = new_element
    assert sample_progression[0] == new_element.ln_id


def test_setitem_slice(sample_progression):
    new_elements = [MockElement(value=f"new_{i}") for i in range(2)]
    sample_progression[0:2] = new_elements
    assert sample_progression[0] == new_elements[0].ln_id
    assert sample_progression[1] == new_elements[1].ln_id


def test_delitem(sample_progression):
    original_length = len(sample_progression)
    del sample_progression[0]
    assert len(sample_progression) == original_length - 1


def test_delitem_slice(sample_progression):
    original_length = len(sample_progression)
    del sample_progression[0:2]
    assert len(sample_progression) == original_length - 2


def test_iter(sample_progression, sample_elements):
    for prog_item, element in zip(sample_progression, sample_elements):
        assert prog_item == element.ln_id


def test_next(sample_progression, sample_elements):
    assert next(sample_progression) == sample_elements[0].ln_id


def test_next_empty():
    empty_prog = Progression()
    with pytest.raises(StopIteration):
        next(empty_prog)


def test_size(sample_progression, sample_elements):
    assert sample_progression.size() == len(sample_elements)


def test_clear(sample_progression):
    sample_progression.clear()
    assert len(sample_progression) == 0


@pytest.mark.parametrize(
    "input_item",
    [
        MockElement(value="new"),
        [MockElement(value="new1"), MockElement(value="new2")],
    ],
)
def test_append(sample_progression, input_item):
    original_length = len(sample_progression)
    sample_progression.append(input_item)
    assert len(sample_progression) > original_length

    if isinstance(input_item, list):
        for item in input_item:
            assert SysUtil.get_id(item) in sample_progression
    else:
        assert SysUtil.get_id(input_item) in sample_progression


def test_pop(sample_progression):
    original_length = len(sample_progression)
    popped_item = sample_progression.pop()
    assert len(sample_progression) == original_length - 1
    assert popped_item not in sample_progression


def test_pop_with_index(sample_progression):
    original_first_item = sample_progression[0]
    popped_item = sample_progression.pop(0)
    assert popped_item == original_first_item
    assert popped_item not in sample_progression


def test_pop_empty():
    empty_prog = Progression()
    with pytest.raises(ItemNotFoundError):
        empty_prog.pop()


@pytest.mark.parametrize(
    "input_item",
    [
        MockElement(value="new"),
        [MockElement(value="new1"), MockElement(value="new2")],
    ],
)
def test_include(sample_progression, input_item):
    original_length = len(sample_progression)
    sample_progression.include(input_item)
    assert len(sample_progression) > original_length

    if isinstance(input_item, list):
        for item in input_item:
            assert SysUtil.get_id(item) in sample_progression
    else:
        assert SysUtil.get_id(input_item) in sample_progression


@pytest.mark.parametrize(
    "input_item",
    [
        MockElement(value="new"),
        [MockElement(value="new1"), MockElement(value="new2")],
    ],
)
def test_exclude(sample_progression, input_item):
    sample_progression.include(input_item)
    original_length = len(sample_progression)
    sample_progression.exclude(input_item)
    assert len(sample_progression) < original_length

    if isinstance(input_item, list):
        for item in input_item:
            assert SysUtil.get_id(item) not in sample_progression
    else:
        assert SysUtil.get_id(input_item) not in sample_progression


def test_is_empty(sample_progression):
    assert not sample_progression.is_empty()
    sample_progression.clear()
    assert sample_progression.is_empty()


def test_remove(sample_progression, sample_elements):
    to_remove = sample_elements[2]
    sample_progression.remove(to_remove)
    assert to_remove not in sample_progression


def test_remove_non_existent(sample_progression):
    with pytest.raises(ItemNotFoundError):
        sample_progression.remove("non_existent_id")


def test_popleft(sample_progression, sample_elements):
    first_element = sample_elements[0]
    popped = sample_progression.popleft()
    assert popped == first_element.ln_id
    assert first_element not in sample_progression


def test_popleft_empty():
    empty_prog = Progression()
    with pytest.raises(ItemNotFoundError):
        empty_prog.popleft()


@pytest.mark.parametrize("size", [10, 100, 1000])
def test_large_progression(size):
    large_prog = Progression(order=[Element() for _ in range(size)])
    assert len(large_prog) == size
    assert large_prog.size() == size


def test_concurrent_operations():
    prog = Progression()

    async def add_items():
        for _ in range(100):
            prog.append(MockElement(value=generate_random_string(10)))
            await asyncio.sleep(0.01)

    async def remove_items():
        for _ in range(50):
            if not prog.is_empty():
                prog.pop()
            await asyncio.sleep(0.02)

    async def run_concurrent():
        await asyncio.gather(add_items(), remove_items())

    asyncio.run(run_concurrent())
    assert 50 <= len(prog) <= 100


def test_progression_with_custom_elements():
    class CustomElement(Element):
        data: dict

    elements = [CustomElement(data={"value": i}) for i in range(5)]
    prog = Progression(order=elements)
    assert len(prog) == 5
    for element in elements:
        assert element in prog


def test_progression_serialization():
    prog = Progression(name="test_prog")
    serialized = prog.to_dict()
    deserialized = Progression.from_dict(serialized)
    assert deserialized == prog
    assert deserialized.name == prog.name


def test_progression_deep_copy():
    import copy

    prog = Progression(name="test_prog")
    prog_copy = copy.deepcopy(prog)
    assert prog == prog_copy
    assert prog is not prog_copy


def test_progression_memory_usage():
    import sys

    small_prog = Progression(order=[Element() for _ in range(10)])
    large_prog = Progression(order=[Element() for _ in range(10000)])
    small_size = sys.getsizeof(small_prog)
    large_size = sys.getsizeof(large_prog)
    assert large_size >= small_size
    # Ensure memory usage grows sub-linearly
    assert large_size <= small_size * 1000


def test_progression_with_async_generator():
    import asyncio

    async def async_gen():
        for i in range(5):
            await asyncio.sleep(0.1)
            yield Element()

    async def run_test():
        p = Progression(order=[i async for i in async_gen()])
        assert len(p) == 5

    asyncio.run(run_test())


def test_progression_index_with_element():
    elements = [MockElement(value=i) for i in range(5)]
    p = Progression(order=elements)
    assert p.index(elements[2]) == 2


def test_progression_count_with_element():
    el1 = Element()
    el2 = Element()
    elements = [el1, el1, el1, el2, el2]
    p = Progression(order=elements)
    assert p.count(elements[1]) == 3


@pytest.mark.parametrize("method", ["append", "include"])
def test_progression_append_include_equivalence(method):
    p1 = Progression()
    p2 = Progression()

    for i in range(5):
        ele = Element()
        getattr(p1, method)(ele)
        getattr(p2, method)(ele)

    assert p1 == p2


def test_progression_remove_with_element():
    elements = [MockElement(value=i) for i in range(5)]
    p = Progression(order=elements)
    p.remove(elements[2])
    assert len(p) == 4
    assert elements[2].ln_id not in p


def test_progression_memory_efficiency():
    import sys

    # Create a progression with a large number of elements
    p = Progression(order=[Element() for _ in range(1000000)])

    # Calculate memory usage
    memory_usage = sys.getsizeof(p) + sum(
        sys.getsizeof(item) for item in p.order
    )

    # Check if memory usage is reasonable (less than 100MB for 1 million elements)
    assert memory_usage < 100 * 1024 * 1024  # 100MB in bytes


def test_progression_serialization_advanced():
    import json

    class ComplexElement(Element):
        data: dict

    p = Progression(
        order=[
            ComplexElement(data={"value": i, "nested": {"x": i * 2}})
            for i in range(5)
        ]
    )

    serialized = json.dumps(p.to_dict())
    deserialized = Progression.from_dict(json.loads(serialized))

    assert len(deserialized) == 5
    assert all(
        isinstance(elem, str) for elem in deserialized
    )  # IDs are strings
    assert p == deserialized


# File: tests/test_progression.py



import atexit
import json
import os
from pathlib import Path
from typing import Any, TypeVar

from lionabc import BaseManager

from lion_core.generic.log import Log
from lion_core.generic.pile import Pile, pile

T = TypeVar("T", bound=Log)


class LogManager(BaseManager):
    """Manages logging operations and persistence."""

    def __init__(
        self,
        logs: Any = None,
        persist_dir: str | None = None,
        persist_path: str | None = None,
        subfolder: str | None = None,
        file_prefix: str | None = None,
    ) -> None:
        """Initialize the LogManager.

        Args:
            logs: Initial logs to manage.
            persist_dir: Directory for log persistence.
            persist_path: Full path for log persistence.
            subfolder: Subfolder within persist_dir.
            file_prefix: Prefix for log files.
        """
        self.logs: Pile[T] = pile(logs or {}, item_type={Log})
        self.persist_dir = persist_dir
        self.persist_path = persist_path
        self.file_prefix = file_prefix
        self.subfolder = subfolder
        atexit.register(self.save_at_exit)

    async def alog(self, log_: Log, /) -> None:
        """Asynchronously add a log to the pile.

        Args:
            log_: The log to add.
        """
        await self.logs.ainclude(log_)

    async def adump(
        self, clear: bool = True, persist_path: str | Path | None = None
    ) -> dict:
        """Asynchronously dump logs and save them.

        Args:
            clear: Whether to clear logs after dumping.
            persist_path: Path to save the dumped logs.

        Returns:
            The dumped log data.
        """
        async with self.logs.async_lock:
            id_ = self.logs[-1].ln_id[:-6]
            data = self.logs.dump(clear=clear)
            self._save(id_=id_, data=data, persist_path=persist_path)
        return data

    def dump(
        self, clear: bool = True, persist_path: str | Path | None = None
    ) -> dict:
        """Dump logs and save them.

        Args:
            clear: Whether to clear logs after dumping.
            persist_path: Path to save the dumped logs.

        Returns:
            The dumped log data.
        """
        id_ = self.logs[-1].ln_id[:-6]
        data = self.logs.dump(clear=clear)
        self._save(id_=id_, data=data, persist_path=persist_path)
        return data

    def _save(
        self, id_: str, data: dict, persist_path: str | Path | None = None
    ) -> None:
        """Save log data to a file.

        Args:
            id_: Identifier for the log file.
            data: Log data to save.
            persist_path: Path to save the log file.
        """
        persist_path = persist_path or self.persist_path

        if not persist_path:
            persist_dir = self.persist_dir or "./data/logs"
            if self.subfolder:
                persist_dir = f"{persist_dir}/{self.subfolder}"
            persist_path = f"{persist_dir}/{self.file_prefix or ''}{id_}.json"

        os.makedirs(os.path.dirname(persist_path), exist_ok=True)

        with open(persist_path, "w") as f:
            json.dump(data, f)

    def load_json(self, persist_path: str | Path | None = None) -> None:
        """Load logs from a JSON file.

        Args:
            persist_path: Path to load the log file from.

        Raises:
            FileNotFoundError: If the log file is not found.
        """
        persist_path = persist_path or self.persist_path
        try:
            with open(persist_path) as f:
                data = json.load(f)
            self.logs = Pile.load(data=data)
        except FileNotFoundError:
            raise FileNotFoundError(f"Log file not found at {persist_path}")

    def save_at_exit(self) -> None:
        """Save logs when exiting the program."""
        if self.logs:
            self.dump(clear=True)


# File: lion_core/log/log_manager.py



from collections.abc import Mapping
from functools import singledispatch
from typing import Any

from lionfuncs import to_dict, to_list

from lion_core.communication.message import RoledMessage
from lion_core.generic.pile import Pile
from lion_core.setting import LionUndefinedType


def validate_message(
    messages, strict=False
) -> list[RoledMessage] | RoledMessage:
    return _validate_message(messages, strict=strict)


@singledispatch
def _validate_message(
    messages: Any, strict=False
) -> list[RoledMessage] | RoledMessage:
    raise NotImplementedError(f"Invalid messages type: {type(messages)}")


@_validate_message.register(LionUndefinedType)
@_validate_message.register(type(None))
def _(messages, strict=False):
    """Handle None or LionUndefined inputs."""
    return []


@_validate_message.register(RoledMessage)
def _(messages, strict=False):
    """Handle RoledMessage inputs."""
    return [messages]


@_validate_message.register(dict)
def _(messages, strict=False):
    """Handle dictionary inputs."""
    try:
        return [RoledMessage.from_dict(messages)]
    except Exception as e:
        if strict:
            raise ValueError(f"Error in creating message object: {e}")
        else:
            return []


@_validate_message.register(str)
def _(messages, strict=False):
    """Handle string inputs."""
    try:
        _d = to_dict(
            messages, str_type="json", fuzzy_parse=True, suppress=True
        )
        if _d:
            return validate_message(_d)
        _d = to_dict(messages, str_type="xml", suppress=True)
        if _d:
            return validate_message(_d)
    except ValueError as e:
        if strict:
            raise ValueError("Error in converting string to dict: ") from e
        else:
            return []


@_validate_message.register(Mapping)
def _(messages, strict=False):
    """Handle mapping inputs."""
    try:
        _d = to_dict(messages)
        return validate_message(_d)
    except Exception as e:
        if strict:
            raise e
        else:
            return []


@_validate_message.register(list)
def _(messages, strict=False):
    try:
        return to_list([validate_message(d) for d in messages], faltten_=True)
    except Exception as e:
        if strict:
            raise e
        return []


@_validate_message.register(Pile)
def _(messages: Pile, strict=False):
    if messages.is_empty():
        return []
    return validate_message(list(messages), strict=strict)



from lion_core.generic.component import Component
from lion_core.generic.exchange import Exchange
from lion_core.generic.flow import Flow
from lion_core.generic.log import Log
from lion_core.generic.node import Node
from lion_core.generic.note import Note
from lion_core.generic.pile import Pile
from lion_core.generic.progression import Progression

__all__ = [
    "Component",
    "Flow",
    "Note",
    "Log",
    "Exchange",
    "Node",
    "Pile",
    "Progression",
]



import asyncio

from lionabc import BaseProcessor, EventStatus
from typing_extensions import override

from lion_core.action.base import ObservableAction


class ActionProcessor(BaseProcessor):
    """
    A processor class for processing action invokations.

    The `ActionProcessor` manages a queue of actions, processing them according
    to the specified capacity and refresh time. It handles the lifecycle of
    actions from enqueuing to processing and stopping. Optionally, it can
    request permission before processing each action.
    """

    observation_type = ObservableAction

    @override
    async def process(self) -> None:
        """
        Processes the work items in the queue.

        Processes items up to the available capacity. Each action is marked as
        `PROCESSING` before execution. After processing, capacity is reset.
        """
        tasks = set()
        prev, next = None, None

        while self.available_capacity > 0 and self.queue.qsize() > 0:
            if prev and prev.status == EventStatus.PENDING:
                next = prev
                await asyncio.sleep(self.refresh_time)
            else:
                next: ObservableAction = await self.dequeue()

            if await self.request_permission(**next.request):
                next.status = EventStatus.PROCESSING
                task = asyncio.create_task(next.invoke())
                tasks.add(task)
            prev = next

        if tasks:
            await asyncio.wait(tasks)
            self.available_capacity = self.capacity


__all__ = ["ActionProcessor"]
# File: lion_core/action/action_processor.py



import inspect
from typing import Any, Literal

from lionabc.exceptions import LionTypeError
from lionfuncs import LN_UNDEFINED
from typing_extensions import override

from lion_core.communication.message import (
    MessageFlag,
    MessageRole,
    RoledMessage,
)
from lion_core.form.base import BaseForm
from lion_core.form.form import Form
from lion_core.generic.note import Note, note


def prepare_request_response_format(request_fields: dict) -> str:
    return (
        "MUST RETURN JSON-PARSEABLE RESPONSE ENCLOSED BY JSON CODE BLOCKS. "
        f"---\n```json\n{request_fields}\n```\n---"
    ).strip()


def _f(idx: str, x: str, /) -> dict[str, Any]:
    """Create an image_url dict for content formatting."""
    return {
        "type": "image_url",
        "image_url": {
            "url": f"data:image/jpeg;base64,{idx}",
            "detail": x,
        },
    }


def format_image_content(
    text_content: str,
    images: list,
    image_detail: Literal["low", "high", "auto"],
) -> dict[str, Any]:
    """Format text content with images for message content."""
    content = [{"type": "text", "text": text_content}]
    content.extend(_f(i, image_detail) for i in images)
    return content


def prepare_instruction_content(
    guidance: str | None = None,
    instruction: str | None = None,
    context: str | dict | list | None = None,
    request_fields: dict | None = None,
    images: str | list | None = None,
    image_detail: Literal["low", "high", "auto"] | None = None,
) -> Note:
    """Prepare the content for an instruction message."""
    out_ = {}
    if guidance:
        out_["guidance"] = guidance

    out_["instruction"] = instruction or "N/A"
    if context:
        if not isinstance(context, str):
            out_["context"] = context
        else:
            out_["context"] = [context]
    if images:
        out_["images"] = images if isinstance(images, list) else [images]
        out_["image_detail"] = image_detail or "low"
    if request_fields:
        out_["request_fields"] = request_fields
        out_["request_response_format"] = prepare_request_response_format(
            request_fields=request_fields
        )

    return note(
        **{k: v for k, v in out_.items() if v not in [None, LN_UNDEFINED]},
    )


class Instruction(RoledMessage):
    """Represents an instruction message in the system."""

    @override
    def __init__(
        self,
        instruction: Any | MessageFlag,
        context: Any | MessageFlag = None,
        guidance: Any | MessageFlag = None,
        images: list | MessageFlag = None,
        sender: Any | MessageFlag = None,
        recipient: Any | MessageFlag = None,
        request_fields: dict | MessageFlag = None,
        image_detail: Literal["low", "high", "auto"] | MessageFlag = None,
        protected_init_params: dict | None = None,
    ) -> None:
        """Initialize an Instruction instance."""
        message_flags = [
            instruction,
            context,
            images,
            sender,
            recipient,
            request_fields,
            image_detail,
        ]

        if all(x == MessageFlag.MESSAGE_LOAD for x in message_flags):
            super().__init__(**protected_init_params)
            return

        if all(x == MessageFlag.MESSAGE_CLONE for x in message_flags):
            super().__init__(role=MessageRole.USER)
            return

        super().__init__(
            role=MessageRole.USER,
            content=prepare_instruction_content(
                guidance=guidance,
                instruction=instruction,
                context=context,
                images=images,
                request_fields=request_fields,
                image_detail=image_detail,
            ),
            sender=sender or "user",
            recipient=recipient or "N/A",
        )

    @property
    def guidance(self):
        """Return the guidance content."""
        return self.content.get(["guidance"], None)

    @property
    def instruction(self):
        """Return the main instruction content."""
        return self.content.get(["instruction"], None)

    def update_images(
        self,
        images: list | str,
        image_detail: Literal["low", "high", "auto"] = None,
    ) -> None:
        """Add new images and update the image detail level."""
        images = images if isinstance(images, list) else [images]
        _ima: list = self.content.get(["images"], [])
        _ima.extend(images)
        self.content["images"] = _ima

        if image_detail:
            self.content["image_detail"] = image_detail

    def update_guidance(self, guidance: str) -> None:
        """Update the guidance content of the instruction."""
        if guidance and isinstance(guidance, str):
            self.content["guidance"] = guidance
            return
        raise LionTypeError(
            "Invalid guidance. Guidance must be a string.",
        )

    def update_request_fields(self, request_fields: dict) -> None:
        """Update the requested fields in the instruction."""
        self.content["request_fields"].update(request_fields)
        format_ = prepare_request_response_format(
            request_fields=self.content["request_fields"],
        )
        self.content["request_response_format"] = format_

    def update_context(self, *args, **kwargs) -> None:
        """Add new context to the instruction."""
        self.content["context"] = self.content.get("context", [])
        if args:
            self.content["context"].extend(args)
        if kwargs:
            self.content["context"].append(kwargs)

    @override
    def _format_content(self) -> dict[str, Any]:
        """Format the content of the instruction."""
        _msg = super()._format_content()
        if isinstance(_msg["content"], str):
            return _msg

        else:
            content = _msg["content"]
            content.pop("images")
            content.pop("image_detail")
            text_content = str(content)
            content = format_image_content(
                text_content=text_content,
                images=self.content.get(["images"]),
                image_detail=self.content.get(["image_detail"]),
            )
            _msg["content"] = content
            return _msg

    @classmethod
    def from_form(
        cls,
        *,
        form: BaseForm | type[Form],
        sender: str | None = None,
        recipient: Any = None,
        images: str | None = None,
        image_detail: str | None = None,
        strict: bool = None,
        assignment: str = None,
        task_description: str = None,
        fill_inputs: bool = True,
        none_as_valid_value: bool = False,
        input_value_kwargs: dict = None,
    ) -> "Instruction":
        """Create an Instruction instance from a form."""
        try:
            if inspect.isclass(form) and issubclass(form, Form):
                form = form(
                    strict=strict,
                    assignment=assignment,
                    task_description=task_description,
                    fill_inputs=fill_inputs,
                    none_as_valid_value=none_as_valid_value,
                    **(input_value_kwargs or {}),
                )

            elif isinstance(form, BaseForm) and not isinstance(form, Form):
                form = Form.from_form(
                    form=form,
                    assignment=assignment or form.assignment,
                    strict=strict,
                    task_description=task_description,
                    fill_inputs=fill_inputs,
                    none_as_valid_value=none_as_valid_value,
                    **(input_value_kwargs or {}),
                )

            if isinstance(form, Form):
                self = cls(
                    guidance=form.guidance,
                    images=images,
                    sender=sender,
                    recipient=recipient,
                    image_detail=image_detail,
                    **form.instruction_dict,
                )
                self.metadata.set(["original_form"], form.ln_id)
                return self
        except Exception as e:
            raise LionTypeError(
                "Invalid form. The form must be a subclass or an instance "
                "of BaseForm."
            ) from e


__all__ = ["Instruction"]

# File: lion_core/communication/instruction.py



# import pytest

# from lionabc import Observable
# from lionabc.exceptions import LionIDError
# from lion_core.setting import LionIDConfig
# from lion_core.sys_utils import SysUtil


# class TestSysUtilID:
#     def test_id_generation_default(self):
#         id_ = SysUtil.id()
#         assert isinstance(id_, str)
#         assert len(id_) == 48  # 42 chars + prefix "ln" + 4 hyphens
#         assert id_.startswith("ln")
#         assert id_.count("-") == 4

#     def test_id_generation_custom_length(self):
#         id_ = SysUtil.id(n=20)
#         assert len(id_) == 26  # 20 chars + prefix "ln" + 4 hyphens

#     def test_id_generation_no_hyphens(self):
#         id_ = SysUtil.id(random_hyphen=False)
#         assert "-" not in id_
#         assert len(id_) == 44  # 42 chars + prefix "ln"

#     def test_id_generation_custom_prefix_postfix(self):
#         id_ = SysUtil.id(prefix="test_", postfix="_end")
#         assert id_.startswith("test_")
#         assert id_.endswith("_end")

#     def test_id_generation_custom_hyphens(self):
#         id_ = SysUtil.id(
#             num_hyphens=2, hyphen_start_index=10, hyphen_end_index=-10
#         )
#         assert id_.count("-") == 2
#         assert "-" not in id_[:10] and "-" not in id_[-10:]

#     @pytest.mark.parametrize("n", [30, 50])
#     def test_id_generation_various_lengths(self, n):
#         id_ = SysUtil.id(n=n)
#         assert len(id_) == n + 6  # n chars + prefix "ln" + 4 hyphens

#     def test_is_id_valid(self):
#         assert SysUtil.is_id(SysUtil.id())
#         assert SysUtil.is_id(
#             "ln8d7274-45a36cc0-8b89af8b-c5-154c41996cf228d8a5"
#         )
#         assert SysUtil.is_id("a" * 32)  # Backward compatibility

#     def test_is_id_invalid(self):
#         assert not SysUtil.is_id("invalid_id")
#         assert not SysUtil.is_id(123)
#         assert not SysUtil.is_id("")
#         assert not SysUtil.is_id(
#             "ln" + "a" * 42
#         )  # Correct length but no hyphens

#     def test_is_id_custom_config(self):
#         custom_config = LionIDConfig(
#             prefix="custom-",
#             n=20,
#             num_hyphens=2,
#             random_hyphen=True,
#             hyphen_start_index=6,
#             hyphen_end_index=-6,
#         )
#         valid_id = "custom-a32eac2f22a-62-a45092f"
#         assert SysUtil.is_id(valid_id, custom_config)
#         assert not SysUtil.is_id(valid_id)  # Should fail with default config

#     def test_get_id_observable(self):
#         class MockObservable(Observable):
#             def __init__(self, ln_id):
#                 self.ln_id = ln_id

#         valid_id = SysUtil.id()
#         mock_obj = MockObservable(valid_id)
#         assert SysUtil.get_id(mock_obj) == valid_id

#     def test_get_id_string(self):
#         valid_id = SysUtil.id()
#         assert SysUtil.get_id(valid_id) == valid_id
#         assert SysUtil.get_id("a" * 32) == "a" * 32  # Backward compatibility

#     def test_get_id_sequence(self):
#         valid_id = SysUtil.id()
#         assert SysUtil.get_id([valid_id]) == valid_id
#         assert SysUtil.get_id((valid_id,)) == valid_id

#     def test_get_id_invalid(self):
#         with pytest.raises(LionIDError):
#             SysUtil.get_id("invalid_id")
#         with pytest.raises(LionIDError):
#             SysUtil.get_id(123)
#         with pytest.raises(LionIDError):
#             SysUtil.get_id([])

#     def test_get_id_custom_config(self):
#         custom_config = LionIDConfig(
#             prefix="custom-",
#             n=20,
#             num_hyphens=2,
#             random_hyphen=True,
#             hyphen_start_index=6,
#             hyphen_end_index=-6,
#         )
#         custom_id = "custom-a32eac2f22a-62-a45092f"
#         assert SysUtil.get_id(custom_id, custom_config) == custom_id
#         with pytest.raises(LionIDError):
#             SysUtil.get_id(custom_id)  # Should fail with default config

#     @pytest.mark.parametrize("invalid_input", [None, [], {}, set()])
#     def test_get_id_edge_cases(self, invalid_input):
#         with pytest.raises(LionIDError):
#             SysUtil.get_id(invalid_input)

#     def test_id_default(self):
#         id_ = SysUtil.id()
#         assert isinstance(id_, str)
#         assert len(id_) == 48  # 42 chars + prefix "ln" + 4 hyphens

#     def test_id_custom_length(self):
#         id_ = SysUtil.id(n=20)
#         assert len(id_) == 26  # 20 chars + prefix "ln" + 4 hyphens

#     def test_id_no_hyphens(self):
#         id_ = SysUtil.id(random_hyphen=False)
#         assert "-" not in id_

#     def test_id_custom_prefix_postfix(self):
#         id_ = SysUtil.id(prefix="test_", postfix="_end")
#         assert id_.startswith("test_")
#         assert id_.endswith("_end")

#     def test_id_all_options(self):
#         id_ = SysUtil.id(
#             n=30,
#             prefix="custom_",
#             postfix="_end",
#             random_hyphen=True,
#             num_hyphens=3,
#         )
#         assert id_.startswith("custom_")
#         assert id_.endswith("_end")
#         assert id_.count("-") == 3
#         assert (
#             len(id_) == 30 + len("custom_") + len("_end") + 3
#         )  # 30 chars + prefix + postfix + 3 hyphens

#     def test_get_id_valid(self):
#         class MockObservable(Observable):
#             def __init__(self, ln_id):
#                 self.ln_id = ln_id

#         valid_id = SysUtil.id()
#         mock_obj = MockObservable(valid_id)
#         assert SysUtil.get_id(mock_obj) == valid_id
#         assert SysUtil.get_id(valid_id) == valid_id

#     def test_get_id_invalid(self):
#         with pytest.raises(LionIDError):
#             SysUtil.get_id("invalid_id")

#     def test_get_id_sequence(self):
#         valid_id = SysUtil.id()
#         assert SysUtil.get_id([valid_id]) == valid_id

#     def test_get_id_custom_config(self):
#         custom_config = LionIDConfig(
#             prefix="custom-",
#             n=20,
#             num_hyphens=2,
#             random_hyphen=True,
#             hyphen_start_index=6,
#             hyphen_end_index=-6,
#         )
#         valid_id = "custom-a32eac2f22a-62-a45092f"
#         assert SysUtil.get_id(valid_id, custom_config) == valid_id

#     def test_is_id_valid(self):
#         valid_id = "ln8d7274-45a36cc0-8b89af8b-c5-154c41996cf228d8a5"
#         assert SysUtil.is_id(valid_id) == True

#     def test_is_id_invalid(self):
#         invalid_id = "invalid_id"
#         assert SysUtil.is_id(invalid_id) == False

#     def test_is_id_custom_config(self):
#         custom_config = LionIDConfig(
#             prefix="custom-",
#             n=20,
#             num_hyphens=2,
#             random_hyphen=True,
#             hyphen_start_index=6,
#             hyphen_end_index=-6,
#         )
#         valid_id = "custom-a32eac2f22a-62-a45092f"
#         assert SysUtil.is_id(valid_id, custom_config) == True

#     def test_get_id_32_char(self):
#         valid_id = "a" * 32
#         assert SysUtil.get_id(valid_id) == valid_id

#     def test_get_id_invalid_type(self):
#         with pytest.raises(LionIDError):
#             SysUtil.get_id(123)

#     def test_is_id_edge_cases(self):
#         assert SysUtil.is_id("a" * 32) == True
#         assert SysUtil.is_id(123) == False
#         assert SysUtil.is_id("") == False



from typing import Any

from lionabc import AbstractElement, Real
from pydantic import BaseModel, Field

from lion_core.sys_utils import SysUtil


class RealElement(BaseModel, AbstractElement, Real):
    """Base class in the Lion framework."""

    ln_id: str = Field(
        default_factory=SysUtil.id,
        title="Lion ID",
        description="Unique identifier for the element",
        frozen=True,
    )

    timestamp: float = Field(
        default_factory=lambda: SysUtil.time(type_="timestamp"),
        title="Creation Timestamp",
        frozen=True,
        alias="created",
    )

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump()

    @classmethod
    def from_dict(cls, data: dict, /, **kwargs) -> "RealElement":
        return cls.model_validate(data, **kwargs)



from lionabc.exceptions import (
    LionOperationError,
    LionTypeError,
    LionValueError,
)

from lion_core.generic.note import note

RESTRICTED_FIELDS = {
    "input_fields",
    "request_fields",
    "init_input_kwargs",
    "output_fields",
}

_err_map = {
    "type": {
        "not_dict": lambda x: LionTypeError(
            message="Input should be a valid dictionary for init.",
            expected_type=dict,
            actual_type=type(x),
        ),
        "not_form_instance": lambda x: LionTypeError(
            message="Invalid form. Should be a instance of Form.",
            expected_type="Form",
            actual_type=type(x),
        ),
        "not_form_class": lambda x: LionTypeError(
            message="Invalid form class, must be a subclass of Form.",
            expected_type="Type[Form]",
            actual_type=type(x),
        ),
        "not_list": lambda x, y: LionTypeError(
            message=y,
            expected_type="list[str]",
            actual_type=type(x),
        ),
    },
    "assignment": {
        "no_assignment": LionValueError(
            "Please provide a valid assignment for this form.",
            "Example assignment: 'input1, input2 -> output'.",
        ),
        "invalid_assignment": lambda x: LionValueError(
            f"Invalid assignment. Field {x} is not found in the form."
        ),
        "explicit_task": LionValueError(
            "Explicitly defining task is not supported. "
            "Please use task_description.",
        ),
        "explcit_input": LionValueError(
            message=(
                "Explicitly defining input_fields is not supported. "
                "Please use assignment to indicate them."
            )
        ),
        "explcit_request": LionValueError(
            message=(
                "Explicitly defining request_fields is not supported. "
                "Please use assignment to indicate them."
            )
        ),
        "strict": lambda x: AttributeError(
            f"The form is set to strict_assignment. {x} "
            "should not be modified after init.",
        ),
        "strict_processed": LionOperationError(
            "The strict form has been processed, and cannot be worked on again"
        ),
        "missing_input": LionValueError(
            "Input fields are missing in the assignment.",
        ),
        "missing_request": LionValueError(
            "Request fields are missing in the assignment.",
        ),
        "incomplete_input": lambda x: LionOperationError(
            message=f"Input fields {x} are not completed.",
        ),
        "incomplete_request": lambda x: LionOperationError(
            message=f"Request fields {x} are not completed.",
        ),
    },
    "field": {
        "missing": lambda x: LionValueError(f"Field {x} is missing."),
        "error": lambda x: LionOperationError(f"Field operation failed: {x}"),
        "modify_input_request_list": LionValueError(
            "input_fields/request_fields list should not be modified"
            "directly. Please use append_to_input/append_to_request"
        ),
        "modify_restricted": lambda x: LionValueError(
            f"{x} should not be modified directly."
        ),
    },
}

ERR_MAP = note(**_err_map)


def get_input_output_fields(str_: str) -> tuple[list[str], list[str]]:
    if str_ is None:
        return [], []

    if "->" not in str_:
        raise ValueError(
            "Invalid assignment format. Expected 'inputs -> outputs'.",
        )

    inputs, outputs = str_.split("->")
    input_fields = [str(i).strip().lower() for i in inputs.split(",")]
    request_fields = [str(o).strip().lower() for o in outputs.split(",")]

    return input_fields, request_fields



from typing import Any

from typing_extensions import override

from lion_core.communication.message import (
    MessageFlag,
    MessageRole,
    RoledMessage,
)


class AssistantResponse(RoledMessage):
    """Represents a response from an assistant in the system."""

    @override
    def __init__(
        self,
        assistant_response: dict | MessageFlag,
        sender: Any | MessageFlag,
        recipient: Any | MessageFlag,
        protected_init_params: dict | None = None,
    ) -> None:
        """Initialize an AssistantResponse instance.

        Args:
            assistant_response: The content of the assistant's response.
            sender: The sender of the response, typically the assistant.
            recipient: The recipient of the response.
            protected_init_params: Optional parameters for protected init.
        """
        message_flags = [assistant_response, sender, recipient]

        if all(x == MessageFlag.MESSAGE_LOAD for x in message_flags):
            super().__init__(**protected_init_params)
            return

        if all(x == MessageFlag.MESSAGE_CLONE for x in message_flags):
            super().__init__(role=MessageRole.ASSISTANT)
            return

        super().__init__(
            role=MessageRole.ASSISTANT,
            sender=sender or "N/A",
            recipient=recipient,
        )
        if assistant_response:
            if isinstance(assistant_response, str):
                assistant_response = {"content": assistant_response}
            elif isinstance(assistant_response, dict):
                if "content" not in assistant_response:
                    assistant_response = {"content": assistant_response}
        else:
            assistant_response = {"content": ""}

        res = assistant_response.get("content", "")
        self.content["assistant_response"] = res

    @property
    def response(self) -> Any:
        """Return the assistant response content."""
        return self.content.get("assistant_response")


# File: lion_core/communication/assistant_response.py



import inspect

from lionabc import BaseRecord
from lionabc.exceptions import LionTypeError, LionValueError
from lionfuncs import to_dict
from pydantic import Field

from lion_core.generic.element import Element
from lion_core.generic.flow import Flow, flow
from lion_core.generic.note import Note, note
from lion_core.generic.pile import Pile, pile
from lion_core.rule.base import Rule
from lion_core.rule.default_rules._default import (
    DEFAULT_RULE_INFO,
    DEFAULT_RULEORDER,
)
from lion_core.sys_utils import SysUtil


class RuleBook(Element, BaseRecord):
    rules_info: Note = note()
    active_rules: Pile[Rule] = Field(
        default_factory=lambda: pile(
            item_type=Rule,
            strict=False,
        ),
        exclude=True,
    )
    default_rule_order: list = Field(default_factory=list)
    rule_flow: Flow = Field(default_factory=lambda: flow({}, "main"))

    def __init__(
        self,
        rules_info=DEFAULT_RULE_INFO,
        default_rule_order=DEFAULT_RULEORDER,
    ):
        super().__init__()
        self.rules_info = validate_rules_info(rules_info)
        self.default_rule_order = default_rule_order or list(
            self.rules_info.keys()
        )
        self.rule_flow.default_name = "main"

    def init_rule(
        self,
        rule: str | type[Rule],
        info: dict | Note = None,
        progress=None,
        **kwargs,
    ):
        if inspect.isclass(rule):
            rule = rule.__name__

        if isinstance(rule, str):
            rule = self.rules_info.get([rule, "rule"], None)

        if rule:
            _info = SysUtil.copy(self.rules_info[rule.__name__])
            _info.pop("rule")
            rule = rule(info=info or _info, **kwargs)
            rule._is_init = True
            self.active_rules.include(rule)
            self.rule_flow.append(rule, progress)
            return rule

        raise LionValueError(f"Invalid rule: {rule}")

    def get_rule_progress(self, progress):
        return self.rule_flow._find_prog(progress, None)


def validate_rules_info(rules_info):
    out = note()

    if isinstance(rules_info, Note):
        rules_info = to_dict(rules_info)["content"]

    if isinstance(rules_info, dict):
        for k, v in rules_info.items():
            v = to_dict(v)
            k: type[Rule] | None = k or v.get("rule", None)
            obj_ = rules_info[k]["rule"]
            if inspect.isclass(obj_) and not issubclass(obj_, Rule):

                raise LionTypeError(
                    message="Item type must be a subclass of Rule.",
                    expected_type=Rule,
                    actual_type=type(k),
                )

            v["rule"] = obj_
            out[k] = v

    return out


__all__ = ["RuleBook"]



from .base import Rule
from .rule_executor import RuleExecutor
from .rule_processor import RuleProcessor
from .rulebook import RuleBook

__all__ = ["Rule", "RuleBook", "RuleProcessor", "RuleExecutor"]



from __future__ import annotations

import asyncio
import threading
from collections.abc import (
    AsyncIterator,
    Callable,
    Iterable,
    Iterator,
    Sequence,
)
from functools import wraps
from typing import Any, Generic, TypeVar

from lionabc import Collective, Observable
from lionabc.exceptions import (
    ItemExistsError,
    ItemNotFoundError,
    LionTypeError,
    LionValueError,
)
from lionfuncs import LN_UNDEFINED, to_list
from pydantic import Field, field_serializer
from typing_extensions import Self, override

from lion_core.generic.element import Element
from lion_core.generic.progression import Progression, prog
from lion_core.generic.utils import to_list_type, validate_order
from lion_core.sys_utils import SysUtil

T = TypeVar("T", bound=Observable)


def synchronized(func: Callable):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        with self.lock:
            return func(self, *args, **kwargs)

    return wrapper


def async_synchronized(func: Callable):
    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        async with self.async_lock:
            return await func(self, *args, **kwargs)

    return wrapper


PILE_KEY_TYPE = int | str | slice


class Pile(Element, Collective, Generic[T]):
    """
    thread-safe async-compatible, ordered collection of Observable elements.

    Pile is a core container in the Lion framework for managing collections
    of Observable objects. It maintains item order and allows fast access
    by unique identifiers, combining list and dictionary characteristics.

    Key Features:
        - Ordered storage with fast access (O(1) by index or Lion ID)
        - Optional type enforcement
        - Thread-safe write operations
        - Asynchronous support for key operations
        - Flexible synchronous and asynchronous iteration

    Thread Safety:
        - State-modifying methods are thread-safe
        - Read-only methods are not synchronized for performance
        - Caution advised when iterating during modifications

    Asynchronous Support:
        - Async versions of state-modifying methods (prefixed with 'a')
        - Asynchronous iteration via __aiter__
        - Read-only methods safe in async contexts without 'await'

    Args:
        items: Initial items for the pile. Defaults to None.
        item_type: Allowed types for items. Defaults to None (any Observable).
        order: Initial order of items. Defaults to None.
        strict: If True, enforces strict type checking. Defaults to False.

    Attributes:
        pile_ (dict[str, T]): Internal storage mapping Lion IDs to items.
        item_type (set[Type[Observable]] | None): Set of allowed item types.
        order (Progression): Maintains the order of items.
        strict (bool): Whether to enforce strict type checking.

    Example:
        >>> from lion_core.generic.pile import Pile
        >>> from lionabc import Observable
        >>>
        >>> class MyItem(Observable):
        ...     def __init__(self, value):
        ...         super().__init__()
        ...         self.value = value
        >>>
        >>> # Create a Pile
        >>> my_pile = Pile(items=[MyItem(1), MyItem(2)], item_type={MyItem})
        >>>
        >>> # Add items
        >>> my_pile.include(MyItem(3))
        >>>
        >>> # Access items
        >>> item = my_pile[0]  # By index
        >>> print(item.value)
        1
        >>>
        >>> # Iterate
        >>> for item in my_pile:
        ...     print(item.value)
        1
        2
        3
        >>>
        >>> # Remove items
        >>> removed_item = my_pile.pop(0)
        >>> print(removed_item.value)
        1

    Notes:
        - Maintains a snapshot view during iteration to prevent concurrent
          modification issues.
        - When subclassing, use @synchronized and @async_synchronized
          decorators for methods that modify state to maintain thread-safety.
        - Not safe for concurrent writes from multiple threads or asyncio
          tasks without external synchronization.

    Warning:
        Modifying the Pile during iteration may lead to unexpected behavior.
        Use the iteration snapshot for safe concurrent access.
    """

    pile_: dict[str, T] = Field(default_factory=dict)
    item_type: set[type[Observable]] | None = Field(
        default=None,
        description="Set of allowed types for items in the pile.",
        exclude=True,
    )
    order: Progression = Field(
        default_factory=prog,
        description="Progression specifying the order of items in the pile.",
        exclude=True,
    )
    strict: bool = Field(
        default=False,
        description="Specify if enforce a strict type check",
    )

    def __pydantic_extra__(self) -> dict[str, Any]:
        return {
            "_lock": Field(default_factory=threading.Lock),
            "_async": Field(default_factory=asyncio.Lock),
        }

    def __pydantic_private__(self) -> dict[str, Any]:
        return self.__pydantic_extra__()

    @override
    def __init__(
        self,
        items: Any = None,
        item_type: set[type[Observable]] | None = None,
        order: Progression | list | None = None,
        strict: bool = False,
        **kwargs,
    ) -> None:
        """
        Initialize a Pile instance.

        Args:
            items: Initial items for the pile.
            item_type: Allowed types for items in the pile.
            order: Initial order of items (as Progression).
            strict: If True, enforce strict type checking.
        """
        _config = {}
        if "ln_id" in kwargs:
            _config["ln_id"] = kwargs["ln_id"]
        if "created" in kwargs:
            _config["created"] = kwargs["created"]

        super().__init__(**_config)
        self.strict = strict
        self.item_type = self._validate_item_type(item_type)
        self.pile_ = self._validate_pile(items or kwargs.get("pile_", {}))
        self.order = self._validate_order(order)

    # Sync Interface methods
    @override
    @classmethod
    def from_dict(
        cls: type[T],
        data: dict[str, Any],
        /,
    ) -> Pile:
        """Create a Pile instance from a dictionary.

        Args:
            data: A dictionary containing Pile data.

        Returns:
            A new Pile instance created from the provided data.

        Raises:
            ValueError: If the dictionary format is invalid.
        """
        items = data.pop("pile_", [])
        items = [Element.from_dict(i) for i in items]
        return cls(items=items, **data)

    def __setitem__(self, key: PILE_KEY_TYPE, item: T | Sequence[T]) -> None:
        """Set an item or items in the Pile.

        Args:
            key: The key to set. Can be an integer index, a string ID, or a
                slice.
            item: The item or items to set. Must be of type T or a sequence
                of T for slices.

        Raises:
            TypeError: If the item type is not allowed.
            KeyError: If the key is invalid.
            ValueError: If trying to set multiple items with a non-slice key.
        """
        self._setitem(key, item)

    @synchronized
    def pop(
        self,
        key: PILE_KEY_TYPE,
        default: Any = LN_UNDEFINED,
        /,
    ) -> T | Pile | None:
        """Remove and return an item or items from the Pile.

        Args:
            key: The key of the item(s) to remove. Can be an integer index,
                a string ID, or a slice.
            default: The value to return if the key is not found. Defaults to
                LN_UNDEFINED.

        Returns:
            The removed item(s), or the default value if not found.

        Raises:
            KeyError: If the key is not found and no default is provided.
        """
        return self._pop(key, default)

    def remove(
        self,
        item: T,
        /,
    ) -> None:
        """Remove a specific item from the Pile.

        Args:
            item: The item to remove.

        Raises:
            ValueError: If the item is not found in the Pile.
        """
        self._remove(item)

    def include(
        self,
        item: T | Iterable[T],
        /,
    ) -> None:
        """Include item(s) in the Pile if not already present.

        Args:
            item: Item or iterable of items to include.

        Raises:
            TypeError: If the item(s) are not of allowed types.
        """
        self._include(item)

    def exclude(
        self,
        item: T | Iterable[T],
        /,
    ) -> None:
        """Exclude item(s) from the Pile if present.

        Args:
            item: Item or iterable of items to exclude.

        Note:
            This method does not raise an error if an item is not found.
        """
        self._exclude(item)

    @synchronized
    def clear(self) -> None:
        """Remove all items from the Pile."""
        self._clear()

    def update(
        self,
        other: Any,
        /,
    ) -> None:
        """Update Pile with items from another iterable or Pile.

        Args:
            other: An iterable or another Pile to update from.

        Raises:
            TypeError: If the items in 'other' are not of allowed types.
        """
        self._update(other)

    @synchronized
    def insert(self, index: int, item: T, /) -> None:
        """Insert an item at a specific position in the Pile.

        Args:
            index: The index at which to insert the item.
            item: The item to insert.

        Raises:
            IndexError: If the index is out of range.
            TypeError: If the item is not of an allowed type.
        """
        self._insert(index, item)

    @synchronized
    def append(self, item: T, /) -> None:
        """Append an item to the end of the Pile.

        This method is an alias for `include`.

        Args:
            item: The item to append.

        Raises:
            TypeError: If the item is not of an allowed type.
        """
        self.update(item)

    @synchronized
    def get(
        self,
        key: PILE_KEY_TYPE,
        default: Any = LN_UNDEFINED,
        /,
    ) -> T | Pile | None:
        """Retrieve item(s) associated with the given key.

        Args:
            key: The key of item(s) to retrieve. Can be an integer index,
                a string ID, or a slice.
            default: Value to return if the key is not found. Defaults to
                LN_UNDEFINED.

        Returns:
            The item(s) associated with the key, or the default value if not
            found. Returns a new Pile instance for slice keys.

        Note:
            Unlike `__getitem__`, this method does not raise KeyError for
            missing keys when a default is provided.
        """
        return self._get(key, default)

    def keys(self) -> Sequence[str]:
        """Return a sequence of all keys (Lion IDs) in the Pile.

        Returns:
            A sequence of string keys representing the Lion IDs of items
            in their current order.
        """
        return list(self.order)

    def values(self) -> Sequence[T]:
        """Return a sequence of all values in the Pile.

        Returns:
            A sequence of all items in the Pile in their current order.
        """
        return [self.pile_[key] for key in self.order]

    def items(self) -> Sequence[tuple[str, T]]:
        """Return a sequence of all (key, value) pairs in the Pile.

        Returns:
            A sequence of tuples, each containing a string key (Lion ID)
            and its corresponding item, in their current order.
        """
        return [(key, self.pile_[key]) for key in self.order]

    def is_empty(self) -> bool:
        """Check if the Pile is empty.

        Returns:
            True if the Pile contains no items, False otherwise.
        """

        return len(self.order) == 0

    def size(self) -> int:
        """Get the number of items in the Pile.

        Returns:
            The count of items currently in the Pile.

        Note:
            This method is equivalent to using the `len()` function
            on the Pile.
        """
        return len(self.order)

    def __iter__(self) -> Iterator[T]:
        """Return an iterator over the items in the Pile.

        This method creates a snapshot of the current order to prevent
        issues with concurrent modifications during iteration.

        Yields:
            Items in the Pile in their current order.
        """
        with self.lock:
            current_order = list(self.order)

        for key in current_order:
            yield self.pile_[key]

    def __next__(self) -> T:
        """Return the next item in the Pile.

        Returns:
            The next item in the Pile.

        Raises:
            StopIteration: When there are no more items in the Pile.
        """
        try:
            return next(iter(self))
        except StopIteration:
            raise StopIteration("End of pile")

    def __getitem__(self, key: PILE_KEY_TYPE) -> Any | list | T:
        """Get item(s) from the Pile by index, ID, or slice.

        Args:
            key: Integer index, string ID, or slice.

        Returns:
            The item or a new Pile containing the sliced items.

        Raises:
            KeyError: If the key is not found.
            TypeError: If the key type is invalid.
        """
        return self._getitem(key)

    def __contains__(self, item: Any) -> bool:
        """Check if an item is in the Pile.

        Args:
            item: The item to check for.

        Returns:
            True if the item is in the Pile, False otherwise.
        """
        return item in self.order

    def __len__(self) -> int:
        """Return the number of items in the Pile.

        Returns:
            The number of items in the Pile.
        """
        return len(self.pile_)

    @override
    def __bool__(self) -> bool:
        """Check if the Pile is not empty.

        Returns:
            True if the Pile is not empty, False otherwise.
        """
        return not self.is_empty()

    def __list__(self) -> list[T]:
        """Convert the Pile to a list.

        Returns:
            A list containing all items in the Pile.
        """
        return self.values()

    def __ior__(self, other: Any | Pile) -> Pile:
        if not isinstance(other, Pile):
            raise LionTypeError(
                "Invalid type for Pile operation.",
                expected_type=Pile,
                actual_type=type(other),
            )
        other = self._validate_pile(list(other))
        self.include(other)
        return self

    def __or__(self, other: Any | Pile) -> Pile:
        if not isinstance(other, Pile):
            raise LionTypeError(
                "Invalid type for Pile operation.",
                expected_type=Pile,
                actual_type=type(other),
            )

        result = self.__class__(
            items=self.values(),
            item_type=self.item_type,
            order=self.order,
        )
        result.include(list(other))
        return result

    def __ixor__(self, other: Any | Pile) -> Pile:
        if not isinstance(other, Pile):
            raise LionTypeError(
                "Invalid type for Pile operation.",
                expected_type=Pile,
                actual_type=type(other),
            )

        to_exclude = []
        for i in other:
            if i in self:
                to_exclude.append(i)

        other = [i for i in other if i not in to_exclude]
        self.exclude(to_exclude)
        self.include(other)
        return self

    def __xor__(self, other: Any | Pile) -> Pile:
        if not isinstance(other, Pile):
            raise LionTypeError(
                "Invalid type for Pile operation.",
                expected_type=Pile,
                actual_type=type(other),
            )

        to_exclude = []
        for i in other:
            if i in self:
                to_exclude.append(i)

        values = [i for i in self if i not in to_exclude] + [
            i for i in other if i not in to_exclude
        ]

        result = self.__class__(
            items=values,
            item_type=self.item_type,
        )
        return result

    def __iand__(self, other: Any | Self) -> Pile:
        if not isinstance(other, Pile):
            raise LionTypeError(
                "Invalid type for Pile operation.",
                expected_type=Pile,
                actual_type=type(other),
            )

        to_exclude = []
        for i in self.values():
            if i not in other:
                to_exclude.append(i)
        self.exclude(to_exclude)
        return self

    def __and__(self, other: Any | Pile) -> Pile:
        if not isinstance(other, Pile):
            raise LionTypeError(
                "Invalid type for Pile operation.",
                expected_type=Pile,
                actual_type=type(other),
            )

        values = [i for i in self if i in other]
        return self.__class__(
            items=values,
            item_type=self.item_type,
        )

    @override
    def __str__(self) -> str:
        """Return a string representation of the Pile.

        Returns:
            A string in the format "Pile(length)".
        """
        return f"Pile({len(self)})"

    @override
    def __repr__(self) -> str:
        """Return a detailed string representation of the Pile.

        Returns:
            A string representation of the Pile, showing its contents
            for small Piles or just the length for larger ones.
        """
        length = len(self)
        if length == 0:
            return "Pile()"
        elif length == 1:
            return f"Pile({next(iter(self.pile_.values())).__repr__()})"
        else:
            return f"Pile({length})"

    def __getstate__(self):
        """Prepare the Pile instance for pickling."""
        state = self.__dict__.copy()
        state["_lock"] = None
        state["_async_lock"] = None
        return state

    def __setstate__(self, state):
        """Restore the Pile instance after unpickling."""
        self.__dict__.update(state)
        self._lock = threading.Lock()
        self._async_lock = asyncio.Lock()

    @property
    def lock(self):
        """Ensure the lock is always available, even during unpickling."""
        if not hasattr(self, "_lock") or self._lock is None:
            self._lock = threading.Lock()
        return self._lock

    @property
    def async_lock(self):
        """Ensure the async lock is always available, even during unpickling"""
        if not hasattr(self, "_async_lock") or self._async_lock is None:
            self._async_lock = asyncio.Lock()
        return self._async_lock

    # Async Interface methods
    @async_synchronized
    async def asetitem(
        self,
        key: PILE_KEY_TYPE,
        item: T | Iterable[T],
        /,
    ) -> None:
        """Asynchronously set an item or items in the Pile.

        Args:
            key: The key to set. Can be an integer index, a string ID, or a
                slice.
            item: The item or items to set. Must be of type T or an iterable
                of T for slices.

        Raises:
            TypeError: If the item type is not allowed.
            KeyError: If the key is invalid.
            ValueError: If trying to set multiple items with a non-slice key.
        """
        self._setitem(key, item)

    @async_synchronized
    async def apop(
        self,
        key: PILE_KEY_TYPE,
        default: Any = LN_UNDEFINED,
        /,
    ):
        """Asynchronously remove and return an item or items from the Pile.

        Args:
            key: The key of the item(s) to remove. Can be an integer index,
                a string ID, or a slice.
            default: The value to return if the key is not found. Defaults to
                LN_UNDEFINED.

        Returns:
            The removed item(s), or the default value if not found.

        Raises:
            KeyError: If the key is not found and no default is provided.
        """
        return self._pop(key, default)

    @async_synchronized
    async def aremove(
        self,
        item: T,
        /,
    ) -> None:
        """Asynchronously remove a specific item from the Pile.

        Args:
            item: The item to remove.

        Raises:
            ValueError: If the item is not found in the Pile.
        """
        self._remove(item)

    @async_synchronized
    async def ainclude(
        self,
        item: T | Iterable[T],
        /,
    ) -> None:
        """Asynchronously include item(s) in the Pile if not already present.

        Args:
            item: Item or iterable of items to include.

        Raises:
            TypeError: If the item(s) are not of allowed types.
        """
        self._include(item)
        if item not in self:
            raise LionTypeError(f"Item {item} is not of allowed types")

    @async_synchronized
    async def aexclude(
        self,
        item: T | Iterable[T],
        /,
    ) -> None:
        """Asynchronously exclude item(s) from the Pile if present.

        Args:
            item: Item or iterable of items to exclude.

        Note:
            This method does not raise an error if an item is not found.
        """
        self._exclude(item)

    @async_synchronized
    async def aclear(self) -> None:
        self._clear()

    @async_synchronized
    async def aupdate(
        self,
        other: Any,
        /,
    ) -> None:
        self._update(other)

    @async_synchronized
    async def aget(
        self,
        key: Any,
        default=LN_UNDEFINED,
        /,
    ) -> list | Any | T:
        return self._get(key, default)

    async def __aiter__(self) -> AsyncIterator[T]:
        """Return an asynchronous iterator over the items in the Pile.

        This method creates a snapshot of the current order to prevent
        issues with concurrent modifications during iteration.

        Yields:
            Items in the Pile in their current order.

        Note:
            This method yields control to the event loop after each item,
            allowing other async operations to run between iterations.
        """

        async with self.async_lock:
            current_order = list(self.order)

        for key in current_order:
            yield self.pile_[key]
            await asyncio.sleep(0)  # Yield control to the event loop

    async def __anext__(self) -> T:
        """Asynchronously return the next item in the Pile."""
        try:
            return await anext(self.AsyncPileIterator(self))
        except StopAsyncIteration:
            raise StopAsyncIteration("End of pile")

    # private methods
    def _getitem(self, key: Any) -> Any | list | T:
        if key is None:
            raise ValueError("getitem key not provided.")

        if isinstance(key, int | slice):
            try:
                result_ids = self.order[key]
                result_ids = (
                    [result_ids]
                    if not isinstance(result_ids, list)
                    else result_ids
                )
                result = []
                for i in result_ids:
                    result.append(self.pile_[i])
                return result[0] if len(result) == 1 else result
            except Exception as e:
                raise ItemNotFoundError(f"index {key}. Error: {e}")

        elif isinstance(key, str):
            try:
                return self.pile_[key]
            except Exception as e:
                raise ItemNotFoundError(f"key {key}. Error: {e}")

        else:
            key = to_list_type(key)
            result = []
            try:
                for k in key:
                    result_id = SysUtil.get_id(k)
                    result.append(self.pile_[result_id])

                if len(result) == 0:
                    raise ItemNotFoundError(f"key {key} item not found")
                if len(result) == 1:
                    return result[0]
                return result
            except Exception as e:
                raise ItemNotFoundError(f"Key {key}. Error:{e}")

    def _setitem(self, key: Any, item: Any) -> None:
        """
        Set new values in the pile using various key types.

        Handles single/multiple assignments, ensures type consistency.
        Supports index/slice, LionID, and LionIDable key access.

        Args:
            key: Key to set items. Can be index, slice, LionID, LionIDable.
            item: Item(s) to set. Can be single item or collection.

        Raises:
            ValueError: Length mismatch or multiple items to single key.
            LionTypeError: Item type doesn't match allowed types.
        """
        item_dict = self._validate_pile(item)

        item_order = []
        for i in item_dict.keys():
            if i in self.order:
                raise ItemExistsError(f"item {i} already exists in the pile")
            item_order.append(i)
        if isinstance(key, int | slice):
            try:
                delete_order = (
                    list(self.order[key])
                    if isinstance(self.order[key], Progression)
                    else [self.order[key]]
                )
                self.order[key] = item_order
                for i in delete_order:
                    self.pile_.pop(i)
                self.pile_.update(item_dict)
            except Exception as e:
                raise ValueError(f"Failed to set pile. Error: {e}")
        else:
            key = to_list_type(key)
            if isinstance(key[0], list):
                key = to_list(key, flatten=True, dropna=True)
            if len(key) != len(item_order):
                raise KeyError(
                    f"Invalid key {key}. Key and item does not match.",
                )
            for k in key:
                id_ = SysUtil.get_id(k)
                if id_ not in item_order:
                    raise KeyError(
                        f"Invalid key {id_}. Key and item does not match.",
                    )
            self.order += key
            self.pile_.update(item_dict)

    def _get(self, key: Any, default: Any = LN_UNDEFINED) -> list | Any | T:
        """
        Retrieve item(s) associated with given key.

        Args:
            key: Key of item(s) to retrieve. Can be single or collection.
            default: Default value if key not found.

        Returns:
            Retrieved item(s) or default if key not found.

        Raises:
            ItemNotFoundError: If key not found and no default specified.
        """
        if isinstance(key, int | slice):
            try:
                return self[key]
            except Exception as e:
                if default is LN_UNDEFINED:
                    raise ItemNotFoundError(f"Item not found. Error: {e}")
                return default
        else:
            check = None
            if isinstance(key, list):
                check = True
                for i in key:
                    if type(i) is not int:
                        check = False
                        break
            try:
                if not check:
                    key = validate_order(key)
                result = []
                for k in key:
                    result.append(self[k])
                if len(result) == 0:
                    raise ItemNotFoundError(f"key {key} item not found")
                if len(result) == 1:
                    return result[0]
                return result

            except Exception as e:
                if default is LN_UNDEFINED:
                    raise ItemNotFoundError(f"Item not found. Error: {e}")
                return default

    def _pop(self, key: Any, default: Any = LN_UNDEFINED) -> list | Any | T:
        """
        Remove and return item(s) associated with given key.

        Args:
            key: Key of item(s) to remove. Can be single or collection.
            default: Default value if key not found.

        Returns:
            Removed item(s) or default if key not found.

        Raises:
            ItemNotFoundError: If key not found and no default specified.
        """
        if isinstance(key, int | slice):
            try:
                pops = self.order[key]
                pops = [pops] if isinstance(pops, str) else pops
                result = []
                for i in pops:
                    self.order.remove(i)
                    result.append(self.pile_.pop(i))
                result = (
                    self.__class__(items=result, item_type=self.item_type)
                    if len(result) > 1
                    else result[0]
                )
                return result
            except Exception as e:
                if default is LN_UNDEFINED:
                    raise ItemNotFoundError(f"Item not found. Error: {e}")
                return default
        else:
            try:
                key = validate_order(key)
                result = []
                for k in key:
                    self.order.remove(k)
                    result.append(self.pile_.pop(k))
                if len(result) == 0:
                    raise ItemNotFoundError(f"key {key} item not found")
                elif len(result) == 1:
                    return result[0]
                return result
            except Exception as e:
                if default is LN_UNDEFINED:
                    raise ItemNotFoundError(f"Item not found. Error: {e}")
                return default

    def _remove(self, item: T):
        """
        Remove an item from the pile.

        Args:
            item: The item to remove.

        Raises:
            ItemNotFoundError: If the item is not found in the pile.
        """
        if item in self:
            self.pop(item)
            return
        raise ItemNotFoundError(f"{item}")

    def _include(self, item: Any):
        """
        Include item(s) in pile if not already present.

        Args:
            item: Item(s) to include. Can be single item or collection.
        """
        item_dict = self._validate_pile(item)

        item_order = []
        for i in item_dict.keys():
            if i not in self.order:
                item_order.append(i)

        self.order.append(item_order)
        self.pile_.update(item_dict)

    def _exclude(self, item: Any):
        """
        Exclude item(s) from pile if present.

        Args:
            item: Item(s) to exclude. Can be single item or collection.
        """
        item = to_list_type(item)
        exclude_list = []
        for i in item:
            if i in self:
                exclude_list.append(i)
        if exclude_list:
            self.pop(exclude_list)

    def _clear(self) -> None:
        """Remove all items from the pile."""
        self.pile_.clear()
        self.order.clear()

    def _update(self, other: Any):
        """Update pile with another collection of items."""
        self.include(other)

    def _validate_item_type(self, value: Any) -> set[type[Observable]] | None:
        """
        Validate the item type for the pile.

        Ensures that the provided item type is a subclass of Element or iModel.
        Raises an error if the validation fails.

        Args:
            value: The item type to validate. Can be a single type or a list of
                    types.

        Returns:
            set: A set of validated item types.

        Raises:
            LionTypeError: If an invalid item type is provided.
            LionValueError: If duplicate item types are detected.
        """
        if value is None:
            return None

        value = to_list_type(value)

        for i in value:
            if not issubclass(i, Observable):
                raise LionTypeError(
                    message="Item type must be a subclass of Observable.",
                    expected_type=Observable,
                    actual_type=type(i),
                )

        if len(value) != len(set(value)):
            raise LionValueError(
                "Detected duplicated item types in item_type.",
            )

        if len(value) > 0:
            return set(value)

    def _validate_pile(self, value: Any) -> dict[str, T]:
        """Validate and convert the items to be added to the pile."""
        if not value:
            return {}

        value = to_list_type(value)

        result = {}
        for i in value:
            if self.item_type:
                if self.strict:
                    if type(i) not in self.item_type:
                        raise LionTypeError(
                            message="Invalid item type in pile."
                            f" Expected {self.item_type}",
                        )
                else:
                    if not any(issubclass(type(i), t) for t in self.item_type):
                        raise LionTypeError(
                            "Invalid item type in pile. Expected "
                            f"{self.item_type} or the subclasses",
                        )
            else:
                if not isinstance(i, Observable):
                    raise LionValueError(f"Invalid pile item {i}")

            result[i.ln_id] = i

        return result

    def _validate_order(self, value: Any) -> Progression:
        if not value:
            return Progression(order=list(self.pile_.keys()))

        if isinstance(value, Progression):
            value = list(value)
        else:
            value = to_list_type(value)

        value_set = set(value)
        if len(value_set) != len(value):
            raise LionValueError("There are duplicate elements in the order")
        if len(value_set) != len(self.pile_.keys()):
            raise LionValueError(
                "The length of the order does not match the length of the pile"
            )

        for i in value_set:
            if SysUtil.get_id(i) not in self.pile_.keys():
                raise LionValueError(
                    f"The order does not match the pile. {i} not found"
                )

        return Progression(order=value)

    def _append(self, item: T):
        """
        Append item to end of pile.

        Args:
            item: Item to append. Can be any lion object, including `Pile`.
        """
        self.update(item)

    def _insert(self, index: int, item: T):
        item_dict = self._validate_pile(item)

        item_order = []
        for i in item_dict.keys():
            if i in self.order:
                raise ItemExistsError(f"item {i} already exists in the pile")
            item_order.append(i)
        self.order.insert(index, item_order)
        self.pile_.update(item_dict)

    @field_serializer("pile_")
    def _(self, value: dict[str, T]):
        return [i.to_dict() for i in value.values()]

    class AsyncPileIterator:
        def __init__(self, pile: Pile):
            self.pile = pile
            self.index = 0

        def __aiter__(self) -> AsyncIterator[T]:
            return self

        async def __anext__(self) -> T:
            if self.index >= len(self.pile):
                raise StopAsyncIteration
            item = self.pile[self.pile.order[self.index]]
            self.index += 1
            await asyncio.sleep(0)  # Yield control to the event loop
            return item

    @async_synchronized
    async def adump(self, clear: bool = False) -> dict:
        self.dump(clear=clear)

    def dump(self, clear: bool = False) -> dict:
        result = self.to_dict()
        if clear:
            self.clear()
        return result

    @classmethod
    def load(cls, data: dict, **kwargs: Any) -> Pile:
        return cls.from_dict(data)


def pile(
    items: Any = None,
    /,
    item_type: type[Observable] | set[type[Observable]] | None = None,
    order: list[str] | None = None,
    strict: bool = False,
    **kwargs,
) -> Pile:
    """
    Create a new Pile instance.

    Args:
        items: Initial items for the pile.
        item_type: Allowed types for items in the pile.
        order: Initial order of items.
        strict: If True, enforce strict type checking.

    Returns:
        Pile: A new Pile instance.
    """

    return Pile(
        items,
        item_type=item_type,
        order=order,
        strict=strict,
        **kwargs,
    )


__all__ = [Pile, Pile]
# File: lion_core/generic/pile.py



from typing import Any

from typing_extensions import override

from lion_core.communication.message import (
    MessageFlag,
    MessageRole,
    RoledMessage,
)
from lion_core.generic.note import Note
from lion_core.sys_utils import SysUtil

DEFAULT_SYSTEM = "You are a helpful AI assistant. Let's think step by step."


def format_system_content(
    system_datetime: bool | str | None,
    system_message: str,
) -> Note:
    """Format the system content with optional datetime information."""
    system_message = system_message or DEFAULT_SYSTEM
    if not system_datetime:
        return Note(system_info=str(system_message))
    if isinstance(system_datetime, str):
        return Note(system_info=f"{system_message}. Date: {system_datetime}")
    if system_datetime:
        date = SysUtil.time(type_="iso", timespec="minutes")
        return Note(system_info=f"{system_message}. System Date: {date}")


class System(RoledMessage):
    """Represents a system message in a language model conversation."""

    @override
    def __init__(
        self,
        system: Any | MessageFlag = None,
        sender: str | None | MessageFlag = None,
        recipient: str | None | MessageFlag = None,
        system_datetime: bool | str | None | MessageFlag = None,
        protected_init_params: dict | None = None,
    ):
        """Initialize a System message instance."""
        if all(
            x == MessageFlag.MESSAGE_LOAD
            for x in [system, sender, recipient, system_datetime]
        ):
            super().__init__(**protected_init_params)
            return

        if all(
            x == MessageFlag.MESSAGE_CLONE
            for x in [system, sender, recipient, system_datetime]
        ):
            super().__init__(role=MessageRole.SYSTEM)
            return

        super().__init__(
            role=MessageRole.SYSTEM,
            sender=sender or "system",
            content=format_system_content(
                system_datetime=system_datetime, system_message=system
            ),
            recipient=recipient or "N/A",
        )

    @property
    def system_info(self) -> Any:
        """Retrieve the system information stored in the message content."""
        return self.content.get("system_info", None)


# File: lion_core/communication/system.py



import pytest
from lionabc.exceptions import LionValueError

from lion_core.communication.action_request import ActionRequest
from lion_core.communication.action_response import (
    ActionResponse,
    prepare_action_response_content,
)
from lion_core.communication.message import MessageFlag, MessageRole
from lion_core.generic.note import Note
from lion_core.sys_utils import SysUtil


# Tests for prepare_action_response_content function
def test_prepare_action_response_content():
    request = ActionRequest(
        "test_func", {"arg1": 1}, SysUtil.id(), SysUtil.id()
    )
    content = prepare_action_response_content(request, "result")
    assert isinstance(content, Note)
    assert content.get("action_request_id") == request.ln_id
    assert content.get(["action_response", "function"]) == "test_func"
    assert content.get(["action_response", "arguments"]) == {"arg1": 1}
    assert content.get(["action_response", "output"]) == "result"


def test_prepare_action_response_content_already_responded():
    request = ActionRequest("test_func", {}, SysUtil.id(), SysUtil.id())
    request.content["action_response_id"] = "existing_response"
    with pytest.raises(LionValueError):
        prepare_action_response_content(request, "result")


# Tests for ActionResponse class
def test_action_response_init():
    request = ActionRequest(
        "test_func", {"arg1": 1}, SysUtil.id(), SysUtil.id()
    )
    response = ActionResponse(request, SysUtil.id(), "result")
    assert response.role == MessageRole.ASSISTANT
    assert response.func_output == "result"
    assert request.is_responded


def test_action_response_init_with_message_load():
    protected_params = {
        "role": MessageRole.ASSISTANT,
        "content": Note(),
        "sender": SysUtil.id(),
    }
    response = ActionResponse(
        MessageFlag.MESSAGE_LOAD,
        MessageFlag.MESSAGE_LOAD,
        MessageFlag.MESSAGE_LOAD,
        protected_init_params=protected_params,
    )
    assert response.role == MessageRole.ASSISTANT


def test_action_response_init_with_message_clone():
    response = ActionResponse(
        MessageFlag.MESSAGE_CLONE,
        MessageFlag.MESSAGE_CLONE,
        MessageFlag.MESSAGE_CLONE,
    )
    assert response.role == MessageRole.ASSISTANT


def test_action_response_func_output():
    request = ActionRequest("test_func", {}, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request, SysUtil.id(), "result")
    assert response.func_output == "result"


def test_action_response_response_dict():
    request = ActionRequest(
        "test_func", {"arg1": 1}, SysUtil.id(), SysUtil.id()
    )
    response = ActionResponse(request, SysUtil.id(), "result")
    assert response.response_dict == {
        "function": "test_func",
        "arguments": {"arg1": 1},
        "output": "result",
    }


def test_action_response_action_request_id():
    request = ActionRequest("test_func", {}, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request, SysUtil.id(), "result")
    assert response.action_request_id == request.ln_id


def test_action_response_update_request():
    request1 = ActionRequest("func1", {"arg1": 1}, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request1, SysUtil.id(), "result1")

    request2 = ActionRequest("func2", {"arg2": 2}, SysUtil.id(), SysUtil.id())
    response.update_request(request2, "result2")

    assert response.func_output == "result2"
    assert response.response_dict["function"] == "func2"
    assert response.response_dict["arguments"] == {"arg2": 2}
    assert request2.is_responded


# Edge cases and additional tests for ActionResponse
def test_action_response_with_none_output():
    request = ActionRequest("test_func", {}, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request, SysUtil.id(), None)
    assert response.func_output is None


def test_action_response_with_large_output():
    request = ActionRequest("test_func", {}, SysUtil.id(), SysUtil.id())
    large_output = "a" * 1000000  # 1MB string
    response = ActionResponse(request, SysUtil.id(), large_output)
    assert len(response.func_output) == 1000000


def test_action_response_unicode():
    request = ActionRequest(
        "test_func", {"arg": "你好世界"}, SysUtil.id(), SysUtil.id()
    )
    response = ActionResponse(request, SysUtil.id(), "こんにちは世界")
    assert response.func_output == "こんにちは世界"


def test_action_response_update_with_different_argument_structure():
    request1 = ActionRequest("func1", {"arg1": 1}, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request1, SysUtil.id(), "result1")

    request2 = ActionRequest(
        "func2", {"arg2": [1, 2, 3]}, SysUtil.id(), SysUtil.id()
    )
    response.update_request(request2, {"complex": "output"})

    assert response.response_dict["arguments"] == {"arg2": [1, 2, 3]}
    assert response.func_output == {"complex": "output"}


def test_action_response_serialization():
    import json

    request = ActionRequest(
        "test_func", {"arg1": 1, "arg2": [1, 2, 3]}, SysUtil.id(), SysUtil.id()
    )
    response = ActionResponse(
        request, SysUtil.id(), {"result": "complex", "data": [4, 5, 6]}
    )

    # Serialize
    response_json = json.dumps(response.to_dict())

    # Deserialize
    reconstructed_response = ActionResponse.from_dict(
        json.loads(response_json)
    )

    assert reconstructed_response.response_dict == response.response_dict


def test_action_response_with_very_deep_nesting():
    def create_nested_dict(depth):
        if depth == 0:
            return {"value": "bottom"}
        return {"nested": create_nested_dict(depth - 1)}

    deep_args = create_nested_dict(100)  # Very deep nesting
    request = ActionRequest("deep_func", deep_args, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request, SysUtil.id(), create_nested_dict(100))

    assert response.func_output == deep_args


def test_action_response_thread_safety():
    import threading

    def create_response():
        request = ActionRequest(
            "thread_func",
            {"thread_id": threading.get_ident()},
            SysUtil.id(),
            SysUtil.id(),
        )
        ActionResponse(
            request,
            SysUtil.id(),
            f"Response from thread {threading.get_ident()}",
        )

    threads = [threading.Thread(target=create_response) for _ in range(100)]
    for t in threads:
        t.start()
    for t in threads:
        t.join()

    # No assertion needed; if this runs without errors, it's thread-safe


def test_action_response_with_invalid_utf8():
    request = ActionRequest("test_func", {}, SysUtil.id(), SysUtil.id())
    invalid_utf8 = b"Invalid UTF-8: \xff"
    with pytest.raises(UnicodeDecodeError):
        ActionResponse(request, SysUtil.id(), invalid_utf8.decode("utf-8"))


def test_action_response_with_no_sender():
    request = ActionRequest("test_func", {}, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request, None, "result")
    assert response.sender == "N/A"


def test_action_response_with_very_long_function_name():
    long_name = "a" * 1000
    request = ActionRequest(long_name, {}, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request, SysUtil.id(), "result")
    assert len(response.response_dict["function"]) == 1000


def test_action_response_with_maximum_arguments():
    max_args = {f"arg{i}": i for i in range(1000)}  # 1000 arguments
    request = ActionRequest(
        "max_args_func", max_args, SysUtil.id(), SysUtil.id()
    )
    response = ActionResponse(request, SysUtil.id(), "result")
    assert len(response.response_dict["arguments"]) == 1000


def test_action_response_update_request_idempotency():
    request1 = ActionRequest("func1", {"arg1": 1}, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request1, SysUtil.id(), "result1")

    request2 = ActionRequest("func2", {"arg2": 2}, SysUtil.id(), SysUtil.id())
    response.update_request(request2, "result2")
    # response.update_request(request2, "result2")  # Update with the same request again

    assert response.func_output == "result2"
    assert response.response_dict["function"] == "func2"
    assert response.response_dict["arguments"] == {"arg2": 2}
    assert request2.is_responded


def test_action_response_with_very_large_request_and_response():
    large_args = {"arg": "a" * 1000000}  # 1MB of arguments
    request = ActionRequest(
        "large_func", large_args, SysUtil.id(), SysUtil.id()
    )
    large_output = "b" * 1000000  # 1MB of output
    response = ActionResponse(request, SysUtil.id(), large_output)

    assert len(response.response_dict["arguments"]["arg"]) == 1000000
    assert len(response.func_output) == 1000000


# Test to ensure that updating a response doesn't affect the original request
def test_action_response_update_request_isolation():
    request1 = ActionRequest("func1", {"arg1": 1}, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request1, SysUtil.id(), "result1")

    request2 = ActionRequest("func2", {"arg2": 2}, SysUtil.id(), SysUtil.id())
    response.update_request(request2, "result2")

    assert request1.request_dict["function"] == "func1"
    assert request1.request_dict["arguments"] == {"arg1": 1}
    assert request1.is_responded


# Test to ensure proper handling of updates with different types
def test_action_response_update_with_different_types():
    request1 = ActionRequest("func1", {"arg1": 1}, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request1, SysUtil.id(), "result1")

    request2 = ActionRequest(
        "func2", {"arg2": "string"}, SysUtil.id(), SysUtil.id()
    )
    response.update_request(request2, 42)

    assert response.func_output == 42
    assert response.response_dict["arguments"] == {"arg2": "string"}


# Test to ensure that the action_request_id is properly updated
def test_action_response_update_action_request_id():
    request1 = ActionRequest("func1", {}, SysUtil.id(), SysUtil.id())
    response = ActionResponse(request1, SysUtil.id(), "result1")

    request2 = ActionRequest("func2", {}, SysUtil.id(), SysUtil.id())
    response.update_request(request2, "result2")

    assert response.action_request_id == request2.ln_id



import pytest
from lionabc.exceptions import LionIDError

from lion_core.generic.component import Component
from lion_core.generic.flow import Flow
from lion_core.generic.progression import prog
from lion_core.sys_utils import SysUtil

sample_progressions = [
    prog([Component(content=i) for i in range(3)], f"prog_{j}")
    for j in range(3)
]

sample_flow = Flow(progressions=sample_progressions, default_name="main")


def test_flow_with_generator_input():
    def gen_progressions():
        for i in range(5):
            yield prog(
                [Component(content=j) for j in range(i)], f"gen_prog_{i}"
            )

    gen_flow = Flow(progressions=list(gen_progressions()))
    assert len(gen_flow) == 5
    assert gen_flow.get("gen_prog_4").size() == 4


@pytest.mark.parametrize(
    "n_progs, prog_size", [(10, 100), (100, 10), (1000, 1)]
)
def test_flow_performance_scaling(n_progs, prog_size):
    import random
    import time

    start_time = time.time()

    large_flow = Flow(
        progressions=[
            prog(
                [Component(content=i) for i in range(prog_size)],
                f"perf_prog_{j}",
            )
            for j in range(n_progs)
        ]
    )

    creation_time = time.time() - start_time

    start_time = time.time()
    for _ in range(100):
        random_prog = random.choice(list(large_flow.keys()))
        _ = large_flow[random_prog]

    access_time = time.time() - start_time

    assert creation_time < 2.0  # Adjusted for larger scale
    assert access_time < 0.5  # Adjusted for larger scale


@pytest.mark.parametrize("op", ["union", "intersection", "difference"])
def test_flow_set_operations(op):
    flow1 = Flow([prog([Component(content=i) for i in range(5)], "set_op_1")])
    flow2 = Flow(
        [prog([Component(content=i) for i in range(3, 8)], "set_op_2")]
    )

    if op == "union":
        result = Flow(
            [
                prog(
                    set(flow1["set_op_1"]) | set(flow2["set_op_2"]),
                    "union",
                )
            ]
        )
        assert result["union"].size() == 10
    elif op == "intersection":
        result = Flow(
            [
                prog(
                    set(flow1["set_op_1"]) & set(flow2["set_op_2"]),
                    "intersection",
                )
            ]
        )
        assert result["intersection"].size() == 0
    elif op == "difference":
        result = Flow(
            [
                prog(
                    set(flow1["set_op_1"]) - set(flow2["set_op_2"]),
                    "difference",
                )
            ]
        )
        assert result["difference"].size() == 5


def test_sys_util_get_id_with_elements():
    element = Component(content=42)
    assert SysUtil.get_id(element) == element.ln_id


def test_sys_util_get_id_with_invalid_input():
    with pytest.raises(LionIDError):
        SysUtil.get_id("not_a_valid_id")


def test_flow_with_proper_id_handling():
    flow = Flow(
        progressions=[prog([], "test_prog_1"), prog([], "test_prog_2")]
    )
    assert SysUtil.is_id(flow.registry["test_prog_1"])
    assert SysUtil.is_id(flow.registry["test_prog_2"])


def test_flow_add_element_with_id():
    flow = Flow([prog([], "test_prog")])
    element = Component(content=100)
    flow.append(element, "test_prog")
    assert SysUtil.get_id(element) in flow["test_prog"].order


# File: tests/test_flow_extended.py
